## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import json
import shutil
import hashlib
import semantic_version
import six
from datetime import datetime
from semantic_version import SpecItem

from . import uhf
uhf(__name__)

from . import center, util, bool_parser

# You have to fill this using https://github.com/workhorsy/py-cpuinfo .
CPU_INFO = None
STABILITES = ('nightly', 'rc', 'stable')


class ModNotFound(Exception):
    mid = None
    spec = None

    def __init__(self, message, mid, spec=None):
        super(ModNotFound, self).__init__(message)

        self.mid = mid
        self.spec = spec


class PackageNotFound(ModNotFound):
    package = None

    def __init__(self, message, mid, package, spec=None):
        super(PackageNotFound, self).__init__(message, mid, spec)

        self.package = package


class NoExecutablesFound(Exception):
    pass


class Repo(object):
    base = None
    is_link = False
    mods = None
    includes = None
    pins = None

    def __init__(self, data=None):
        self.mods = {}
        self.pins = {}

        if data is not None:
            self.set(data)

    def empty(self):
        return len(self.mods) == 0

    def clear(self):
        self.mods = {}
        self.pins = {}

    def load_json(self, path):
        self.base = os.path.dirname(path)
        with open(path, 'r') as stream:
            self.set(json.load(stream))

    def save_json(self, path):
        with open(path, 'w') as stream:
            json.dump(self.get(), stream)

    def set(self, info):
        for m, v in info['pins'].items():
            self.pins[m] = semantic_version.Version(v)

        for mod in info['mods']:
            self.add_mod(Mod(mod, self))

    def get(self):
        mods = []
        for v in self.mods.values():
            for mod in v:
                mods.append(mod.get())

        pins = self.pins.copy()
        for m, v in pins.items():
            pins[m] = str(v)

        return {
            'pins': pins,
            'mods': mods
        }

    def fetch(self, link):
        self.base = os.path.dirname(link)
        self.is_link = True
        self.parse(util.get(link))

    def read(self, path):
        self.base = os.path.dirname(path)
        h = open(path, 'r')
        self.parse(h)
        h.close()

    def parse(self, obj):
        if not obj:
            return

        if isinstance(obj, six.string_types):
            data = json.loads(obj)
        else:
            data = json.load(obj)

        self.mods = {}
        self.includes = data.get('includes', [])

        for inc in self.includes:
            item = Repo()
            if self.is_link:
                item.fetch(util.url_join(self.base, inc))
            else:
                item.read(os.path.join(self.base, inc))

            self.merge(item)

        for mod in data.get('mods', []):
            self.add_mod(Mod(mod, self))

    def add_mod(self, mod):
        mid = mod.mid

        if len(mod.packages) == 0 and not isinstance(mod, InstalledMod):
            logging.warning('Mod %s is empty, ignoring it!', mod)
            return

        if mid in self.mods:
            inserted = False

            for i, item in enumerate(self.mods[mid]):
                if item.version == mod.version:
                    if mod._repo is None:
                        mod_base = 'None'
                    else:
                        mod_base = mod._repo.base

                    logging.info('Mod "%s" (%s) from "%s" overwrites an existing mod version!', mid, mod.version, mod_base)

                    self.mods[mid][i] = mod
                    inserted = True
                    break

            if not inserted:
                self.mods[mid].append(mod)
                self.mods[mid].sort(key=lambda m: m.version, reverse=True)
        else:
            self.mods[mid] = [mod]

        mod._repo = self

    def remove_mod(self, mod):
        mid = mod.mid

        if mid not in self.mods:
            raise ModNotFound('Mod "%s" (%s) could not be removed from %s!' % (mid, mod.version, self.base), mid=mid)

        idx = None
        for i, item in enumerate(self.mods[mid]):
            if item.version == mod.version:
                idx = i
                break

        if not idx:
            raise ModNotFound('Mod "%s" (%s) could not be removed from %s because the exact version was missing!' % (mid, mod.version, self.base), mid=mid)

        del self.mods[mid][idx]

        if len(self.mods[mid]) == 0:
            del self.mods[mid]

    def merge(self, repo):
        for mvs in repo.mods.values():
            for mod in mvs:
                self.add_mod(mod)

    def pin(self, mod, version=None):
        if isinstance(mod, Package):
            mod = mod.get_mod()

        if isinstance(mod, Mod):
            version = mod.version
            mod = mod.mid

        if not isinstance(mod, str) or not isinstance(version, semantic_version.Version):
            raise ValueError('%s is not a string or %s is not a valid version!' % (mod, version))

        self.pins[mod] = version

    def unpin(self, mod):
        if isinstance(mod, Package):
            mod = mod.get_mod().mid
        elif isinstance(mod, Mod):
            mod = mod.mid

        if mod in self.pins:
            del self.pins[mod]

    def get_pin(self, mod):
        if isinstance(mod, Package):
            mod = mod.get_mod()

        if isinstance(mod, Mod):
            mod = mod.mid

        return self.pins.get(mod, None)

    def query(self, mid, spec=None, pname=None):
        if isinstance(mid, Package):
            mod = mid.get_mod()
            pname = mid.name
            mid = mod.mid
            spec = util.Spec.from_version(mod.version)

            del mod
        elif isinstance(mid, Mod):
            spec = util.Spec.from_version(mid.version)
            mid = mid.mid

        if mid not in self.mods:
            raise ModNotFound('Mod "%s" wasn\'t found!' % (mid), mid)

        if spec is None and mid in self.pins:
            spec = util.Spec.from_version(self.pins[mid])

        candidates = self.mods[mid]
        if spec is None:
            mod = candidates[0]
        else:
            if isinstance(spec, semantic_version.Version):
                logging.warning('Repo.query(): Expected Spec but got Version instead! (%s)' % repr(spec))
                spec = util.Spec.from_version(spec)

            if candidates[0].mtype == 'engine':
                # Multiple versions qualify and this is an engine so we have to check the stability next
                stab = center.settings['engine_stability']
                if stab not in STABILITES:
                    stab = STABILITES[-1]

                stab_idx = STABILITES.index(stab)
                version = None
                while stab_idx > -1:
                    version = spec.select([mod.version for mod in candidates if mod.stability == stab])
                    if not version:
                        # Nothing found, try the next lower stability
                        stab_idx -= 1
                        stab = STABILITES[stab_idx]
                    else:
                        # Found at least one result
                        break

                if not version:
                    version = spec.select([mod.version for mod in candidates])
            else:
                version = spec.select([mod.version for mod in candidates])

            if not version:
                raise ModNotFound('Mod "%s" %s wasn\'t found!' % (mid, spec), mid, spec)

            for m in candidates:
                if m.version == version:
                    mod = m
                    break

        if pname is not None:
            for pkg in mod.packages:
                if pkg.name == pname:
                    return pkg

            raise ModNotFound('The package "%s" for mod "%s" wasn\'t found!' % (pname, mid), mid)
        else:
            return mod

    def query_all(self, mid, spec=None):
        if mid not in self.mods:
            raise ModNotFound('Mod "%s" wasn\'t found!' % (mid), mid)

        for mod in self.mods[mid]:
            if spec is None or spec.match(mod.version):
                yield mod

    def has(self, mid, spec=None, pname=None):
        try:
            self.query(mid, spec, pname)
            return True
        except ModNotFound:
            return False

    def get_tree(self):
        roots = set(self.mods.keys())
        return [self.mods[mid][0] for mid in roots]

    def get_list(self):
        for mod in self.mods.values():
            yield mod[0]

    # TODO: Is this overcomplicated?
    def process_pkg_selection(self, pkgs, recursive=True):
        dep_dict = {}
        ndeps = pkgs

        for pkg in pkgs:
            mod = pkg.get_mod()
            dd = dep_dict.setdefault(mod.mid, {})

            version = str(mod.version)
            if version != '*' and not SpecItem.re_spec.match(version):
                # Make a spec out of this version
                version = '==' + version

            dd = dd.setdefault(util.Spec(version), {})
            dd['#mod'] = pkg.get_mod()
            dd[pkg.name] = pkg

        # Resolve the pkgs' dependencies
        while len(ndeps) > 0:
            _nd = ndeps
            ndeps = []

            for pkg in _nd:
                for dep, version in pkg.resolve_deps():
                    dd = dep_dict.setdefault(dep.get_mod().mid, {})
                    dd = dd.setdefault(version, {})
                    dd['#mod'] = dep.get_mod()

                    if dep.name not in dd:
                        dd[dep.name] = dep
                        if recursive:
                            ndeps.append(dep)

        # Check for conflicts (and try to resolve them if possible).
        dep_list = set()
        pref_stable = center.settings['engine_stability']
        for mid, variants in dep_dict.items():
            if len(variants) == 1:
                # This loop only iterates once but it's the easiest solution to get the actual value
                for pkgs in variants.values():
                    del pkgs['#mod']
                    dep_list |= set(pkgs.values())
            else:
                specs = variants.keys()
                remains = []

                for v in variants.values():
                    ok = True
                    for spec in specs:
                        if not spec.match(v['#mod'].version):
                            ok = False
                            break

                    if ok:
                        remains.append(v)

                if len(remains) == 0:
                    v = (list(variants.values())[0]['#mod'].title, ','.join([str(s) for s in specs]))
                    raise PackageNotFound('No version of mod "%s" found for these constraints: %s' % v, mid, v[0])
                else:
                    if remains[0]['#mod'].mtype == 'engine':
                        # Multiple versions qualify and this is an engine so we have to check the stability next
                        stab = pref_stable
                        if stab not in STABILITES:
                            stab = STABILITES[-1]

                        stab_idx = STABILITES.index(stab)
                        candidates = []
                        while stab_idx > -1:
                            candidates = [m for m in remains if m['#mod'].stability == stab]
                            if len(candidates) == 0:
                                # Nothing found, try the next lower stability
                                stab_idx -= 1
                                stab = STABILITES[stab_idx]
                            else:
                                # Found at least one result
                                break

                        # An empty remains list would trigger an index out of bounds error; avoid that
                        if len(candidates) > 0:
                            remains = candidates

                    # Pick the latest
                    remains.sort(key=lambda v: v['#mod'].version)
                    del remains[-1]['#mod']
                    dep_list |= set(remains[-1].values())

        return dep_list

    def get_dependents(self, pkgs):
        deps = set()
        mids = set([p.get_mod().mid for p in pkgs])

        for mid, mvs in self.mods.items():
            # Ignore self-references
            if mid not in mids:
                for mod in mvs:
                    try:
                        for dp in self.process_pkg_selection(mod.packages, False):
                            if dp in pkgs:
                                deps.add(mod)
                    except ModNotFound:
                        pass

        return deps


class Mod(object):
    _repo = None
    mid = ''
    title = ''
    mtype = 'mod'
    version = None
    stability = None
    parent = None
    cmdline = ''
    mod_flag = None
    logo = None
    tile = None
    banner = None
    description = ''
    notes = ''
    release_thread = None
    videos = None
    screenshots = None
    attachments = None
    first_release = None
    last_update = None
    actions = None
    packages = None

    __fields__ = ('mid', 'title', 'type', 'version', 'parent', 'cmdline', 'logo', 'tile', 'banner',
        'description', 'notes', 'actions', 'packages')

    def __init__(self, values=None, repo=None):
        self.actions = []
        self.packages = []
        self.videos = []

        if repo is not None:
            self._repo = repo

        if values is not None:
            self.set(values)

    def __repr__(self):
        return '<Mod "%s" %s (%s)>' % (self.title, self.version, self.mid)

    def set(self, values):
        self.mid = values['id']
        self.title = values['title']
        self.mtype = values.get('type', 'mod')  # Backwards compatibility
        self.version = semantic_version.Version(values['version'], partial=True)
        self.stability = values.get('stability', 'stable')
        self.parent = values.get('parent', 'FS2')
        self.cmdline = values.get('cmdline', '')
        self.mod_flag = values.get('mod_flag', None)
        self.description = values.get('description', '')
        self.notes = values.get('notes', '')
        self.release_thread = values.get('release_thread', None)
        self.videos = values.get('videos', [])
        self.screenshots = values.get('screenshots', [])
        self.attachments = values.get('attachments', [])
        self.first_release = values.get('first_release', None)
        self.last_update = values.get('last_update', None)
        self.actions = values.get('actions', [])

        self.packages = []

        installed = isinstance(self, InstalledMod)
        for pkg in values.get('packages', []):
            p = Package(pkg, self)
            if installed or p.check_env():
                self.packages.append(p)

        base = None
        if hasattr(self, 'folder'):
            base = self.folder
        elif self._repo is not None:
            base = self._repo.base

        if base:
            for prop in ('logo', 'tile', 'banner'):
                if values.get(prop) is not None:
                    if '://' in base:
                        setattr(self, prop, util.url_join(base, values[prop]))
                    elif '://' not in values[prop]:
                        setattr(self, prop, os.path.abspath(os.path.join(base, values[prop])))
                    else:
                        setattr(self, prop, values[prop])

            for prop in ('screenshots', 'attachments'):
                ims = getattr(self, prop)
                for i, path in enumerate(ims):
                    if '://' in base:
                        ims[i] = util.url_join(base, path)
                    elif '://' not in path:
                        ims[i] = os.path.abspath(os.path.join(base, path))

        else:
            for prop in ('logo', 'tile', 'banner'):
                setattr(self, prop, values.get(prop))

        if self.first_release:
            self.first_release = datetime.strptime(self.first_release, '%Y-%m-%d')

        if self.last_update:
            self.last_update = datetime.strptime(self.last_update, '%Y-%m-%d')

        # Enforce relative paths
        for act in self.actions:
            if 'paths' in act:
                act['paths'] = [p.lstrip('/') for p in act['paths']]

            if 'dest' in act:
                act['dest'] = act['dest'].lstrip('/')

    def get(self):
        return {
            'id': self.mid,
            'title': self.title,
            'type': self.mtype,
            'version': str(self.version),
            'stability': self.stability,
            'parent': self.parent,
            'cmdline': self.cmdline,
            'mod_flag': self.mod_flag,
            'logo': self.logo,
            'tile': self.tile,
            'banner': self.banner,
            'description': self.description,
            'notes': self.notes,
            'release_thread': self.release_thread,
            'videos': self.videos,
            'screenshots': self.screenshots,
            'attachments': self.attachments,
            'first_release': self.first_release.strftime('%Y-%m-%d') if self.first_release else None,
            'last_update': self.last_update.strftime('%Y-%m-%d') if self.last_update else None,
            'actions': self.actions,
            'packages': [pkg.get() for pkg in self.packages]
        }

    def copy(self):
        return self.__class__(self.get(), self._repo)

    def get_files(self):
        files = []
        for pkg in self.packages:
            for item in pkg.filelist:
                item = item.copy()
                item['package'] = pkg.name
                files.append(item)

        return files

    def resolve_deps(self, only_required=True, recursive=True):
        if only_required:
            pkgs = [pkg for pkg in self.packages if pkg.status == 'required']
        else:
            pkgs = [pkg for pkg in self.packages if pkg.status in ('required', 'recommended')]

        return self._repo.process_pkg_selection(pkgs, recursive=recursive)

    def get_parent(self):
        if self.parent:
            return self._repo.query(self.parent)
        else:
            return None

    def get_dependents(self):
        return self._repo.get_dependents(self.packages)


class Package(object):
    _mod = None
    name = ''
    notes = ''
    status = 'recommended'
    dependencies = None
    environment = None
    folder = None
    is_vp = False
    files = None
    filelist = None
    executables = None

    def __init__(self, values=None, mod=None):
        self._mod = mod
        self.dependencies = []
        self.environment = []
        self.files = {}

        if values is not None:
            self.set(values)

    def __repr__(self):
        return '<Package "%s" of %s>' % (self.name, self._mod)

    def set(self, values):
        self.name = values['name']
        self.notes = values.get('notes', '')
        self.status = values.get('status', 'recommended').lower()
        self.dependencies = values.get('dependencies', [])
        self.environment = values.get('environment', '')
        self.folder = values.get('folder', self.name)
        self.is_vp = values.get('is_vp', False)
        self.files = {}
        self.filelist = values.get('filelist', [])
        self.executables = []

        if self.folder is None:
            self.folder = self.name

        for exe in values.get('executables', []):
            self.executables.append({
                'file': exe['file'],
                'label': exe.get('label', None)
            })

        _files = values.get('files', [])

        if isinstance(_files, dict):
            self.files = _files
            for name, item in _files.items():
                item['filename'] = name
                item['dest'] = item.get('dest', '').strip('/')  # make sure this is a relative path

        elif isinstance(_files, list):
            for item in _files:
                item['dest'] = item.get('dest', '').strip('/')  # make sure this is a relative path
                self.files[item['filename']] = item
        else:
            logging.warning('"%s"\'s file list has an unknown type.', self.name)

        if self._mod:
            mid = self._mod.mid
            if mid == '':
                raise Exception('Package "%s" initialized with Mod %s which has no ID!' % (self.name, self._mod.title))

    def get(self):
        return {
            'name': self.name,
            'notes': self.notes,
            'status': self.status,
            'dependencies': self.dependencies,
            'environment': self.environment,
            'folder': self.folder,
            'is_vp': self.is_vp,
            'files': list(self.files.values()),
            'filelist': self.filelist,
            'executables': self.executables
        }

    def get_mod(self):
        return self._mod

    def get_files(self):
        files = {}
        for name, item in self.files.items():
            for path, csum in item['contents'].items():
                files[os.path.join(item['dest'], path)] = (csum, name)

        return files

    def resolve_deps(self):
        result = []
        for dep in self.dependencies:
            version = dep.get('version', '*') or '*'
            if version != '*' and not SpecItem.re_spec.match(version):
                # Make a spec out of this version
                version = '==' + version

            version = util.Spec(version)
            mod = self._mod._repo.query(dep['id'], version)
            pkgs = dep.get('packages', [])
            found_pkgs = []

            for pkg in mod.packages:
                if pkg.status == 'required' or pkg.name in pkgs:
                    result.append((pkg, version))
                    found_pkgs.append(pkg.name)

            missing_pkgs = set(pkgs) - set(found_pkgs)
            if len(missing_pkgs) > 0:
                raise PackageNotFound('Package %s of mod %s (%s) couldn\'t be found!' % (next(iter(missing_pkgs)), mod.mid, version), mod.mid, missing_pkgs)

        return result

    def check_env(self):
        if self.environment in ('', None):
            return True

        if not isinstance(self.environment, str):
            logging.warning('Invalid value for environment check in mod %s (%s)!' % (self._mod.mid, self._mod.version))
            return True

        bvars = {}
        bvars[CPU_INFO['arch'].lower()] = True  # this is either X86_32 or X86_64

        if sys.platform in ('win32', 'cygwin'):
            bvars['windows'] = True
        elif sys.platform.startswith('linux'):
            bvars['linux'] = True
        elif sys.platform == 'darwin':
            bvars['macosx'] = True
        else:
            logging.error('You are using an unrecognized OS! (%s)' % sys.platform)

        for flag in CPU_INFO['flags']:
            bvars[flag.lower()] = True

        try:
            return bool_parser.eval_string(self.environment, bvars)
        except Exception:
            logging.exception('Failed to evaluate expression "%s"!' % self.environment)

            # Since we can't perform this check, just assume that it returns True as otherwise this mod would be
            # inaccessible.
            return True


# Keeps track of installed mods
class InstalledRepo(Repo):
    base = '[INSTALLED]'

    def clear(self):
        self.mods = {}

    def save_pins(self, path):
        with open(path, 'w') as stream:
            json.dump(self.pins, stream)

    def load_pins(self, path):
        with open(path, 'r') as stream:
            self.pins = json.load(stream)

    def set(self, mods):
        for m, v in mods['pins'].items():
            self.pins[m] = semantic_version.Version(v)

        for mod in mods['mods']:
            self.add_mod(InstalledMod(mod))

    def add_pkg(self, pkg):
        mod = pkg.get_mod()
        try:
            my_mod = self.query(mod)
        except ModNotFound:
            my_mod = InstalledMod.convert(mod)
            my_pkg = my_mod.add_pkg(pkg)
            self.add_mod(my_mod)
            return my_pkg
        else:
            return my_mod.add_pkg(pkg)

    def del_pkg(self, pkg):
        mod = pkg.get_mod()

        try:
            my_mod = self.query(mod)
        except ModNotFound:
            logging.error('Tried to delete non-existing package!')
            return

        my_mod.del_pkg(pkg)

    def del_mod(self, mod):
        if mod.mid in self.mods:
            vs = self.mods[mod.mid]
            rem = None
            for m in vs:
                if m.version == mod.version:
                    rem = m
                    break

            if rem is None:
                logging.error('Tried to delete missing mod version!')
            else:
                vs.remove(rem)
                if len(vs) == 0:
                    del self.mods[mod.mid]

    def is_installed(self, mid, spec=None, pname=None):
        try:
            self.query(mid, spec, pname)
            return True
        except ModNotFound:
            return False

    def get_updates(self):
        remote_mods = center.mods
        updates = {}

        for mid, mods in self.mods.items():
            try:
                rem_mod = remote_mods.query(mid)
            except ModNotFound:
                continue

            if rem_mod.version > mods[0].version:
                # Let's see if the files changed.
                my_files = {}
                rem_files = {}
                my_pkgs = [pkg.name for pkg in mods[0].packages]

                for item in mods[0].get_files():
                    my_files[item['filename']] = item['checksum']

                for item in rem_mod.get_files():
                    if item['package'] in my_pkgs:
                        rem_files[item['filename']] = item['checksum']

                if rem_files == my_files:
                    logging.warning('Detected an empty update for mod "%s"! (%s -> %s)', mods[0].title, str(mods[0].version), str(rem_mod.version))
                    # TODO: Resolve this situation! (Update the local metadata?)
                else:
                    if mid not in updates:
                        updates[mid] = {}

                    updates[mid][mods[0].version] = rem_mod.version

        return updates


class InstalledMod(Mod):
    check_notes = ''
    folder = None
    dev_mode = False
    custom_build = None
    user_exe = None
    user_cmdline = None
    user_custom_build = None
    _path = None

    @staticmethod
    def load(path):
        if path.endswith('.json'):
            with open(path, 'r') as stream:
                data = json.load(stream)

            mod = InstalledMod(None)
            mod.folder = os.path.normpath(os.path.dirname(path))
            mod.set(data)

            user_path = os.path.join(os.path.dirname(path), 'user.json')
            if os.path.isfile(user_path):
                try:
                    with open(user_path, 'r') as stream:
                        mod.set_user(json.load(stream))
                except Exception:
                    logging.exception('Failed to load user data for %s!' % mod)

            return mod
        else:
            return None

    @staticmethod
    def convert(mod):
        data = mod.get()
        data['packages'] = []

        nmod = InstalledMod(data)
        nmod.generate_folder()
        return nmod

    def set(self, values):
        pkgs = values.get('packages', [])
        values = values.copy()
        values['packages'] = []

        super(InstalledMod, self).set(values)

        if 'folder' in values:
            self.folder = values['folder']

        self.dev_mode = values.get('dev_mode', False)
        self.custom_build = values.get('custom_build', None)
        self.check_notes = values.get('check_notes', '')

        if not self.mod_flag:
            # Fix broken metadata
            self.update_mod_flag()

        for pkg in pkgs:
            installed_pkg = InstalledPackage(pkg, self)
            # If the user installed packages on multiple platforms into the same directory then an installed package
            # may be present that is not valid for the current environment so we need to check that here
            if installed_pkg.check_env():
                self.packages.append(installed_pkg)

    def set_user(self, values):
        self.user_exe = values.get('exe')
        self.user_cmdline = values.get('cmdline')
        self.user_custom_build = values.get('custom_build')

    def get(self):
        return {
            'installed': True,
            'id': self.mid,
            'title': self.title,
            'type': self.mtype,
            'parent': self.parent,
            'version': str(self.version),
            'stability': self.stability,
            'description': self.description,
            'notes': self.notes,
            'folder': self.folder,
            'logo': self.logo,
            'tile': self.tile,
            'banner': self.banner,
            'release_thread': self.release_thread,
            'videos': self.videos,
            'screenshots': self.screenshots[:],
            'attachments': self.attachments[:],
            'first_release': self.first_release.strftime('%Y-%m-%d') if self.first_release else None,
            'last_update': self.last_update.strftime('%Y-%m-%d') if self.last_update else None,
            'cmdline': self.cmdline,
            'mod_flag': self.mod_flag,
            'dev_mode': self.dev_mode,
            'custom_build': self.custom_build,
            'packages': [pkg.get() for pkg in self.packages],

            'user_exe': self.user_exe,
            'user_cmdline': self.user_cmdline,
            'user_custom_build': self.user_custom_build
        }

    def get_user(self):
        return {
            'exe': self.user_exe,
            'cmdline': self.user_cmdline,
            'custom_build': self.user_custom_build
        }

    def get_relative(self):
        info = self.get()

        # Storing the folder of the JSON file inside the file would be silly.
        del info['folder']
        del info['user_exe']
        del info['user_cmdline']
        del info['user_custom_build']

        # Make sure we only store relative paths.
        for prop in ('logo', 'tile', 'banner'):
            if info[prop] and '://' not in info[prop]:
                info[prop] = os.path.relpath(info[prop], self.folder)

        for prop in ('screenshots', 'attachments'):
            paths = info[prop]
            for i, p in enumerate(paths):
                paths[i] = os.path.relpath(p, self.folder)

        return info

    def copy(self):
        return InstalledMod(self.get(), self._repo)

    def generate_folder(self):
        # IMPORTANT: This code decides where newly installed mods are stored.
        base = os.path.normpath(center.settings['base_path'])
        meta = self.get_relative()

        if self.mtype in ('engine', 'tool'):
            self.folder = os.path.join(base, 'bin', self.mid)
        elif self.mtype == 'tc':
            self.folder = os.path.join(base, self.mid, self.mid)
        else:
            self.folder = os.path.join(base, self.parent, self.mid)

        self.folder += '-' + str(self.version)

        # This should update the paths to images, etc. since get_relative() only returns relative paths
        self.set(meta)

    def add_pkg(self, pkg):
        pkg = InstalledPackage.convert(pkg, self)
        found = False

        for i, p in enumerate(self.packages):
            if p.name == pkg.name:
                self.packages[i] = pkg
                found = True
                break

        if not found:
            self.packages.append(pkg)

        return pkg

    def del_pkg(self, pkg):
        for i, p in enumerate(self.packages):
            if p.name == pkg.name:
                del self.packages[i]
                break

    def save(self):
        im_path = util.ipath(self.folder)

        # Correct the casing of our folder if neccessary.
        if im_path != self.folder:
            self.folder = im_path

        with open(os.path.join(self.folder, 'mod.json'), 'w') as stream:
            json.dump(self.get_relative(), stream, indent=4)

    def save_user(self):
        modpath = self.folder
        im_path = util.ipath(modpath)

        # Correct the casing of our folder if neccessary.
        if im_path != modpath:
            modpath = im_path
            self.folder = modpath

        path = os.path.join(modpath, 'user.json')

        with open(path, 'w') as stream:
            json.dump(self.get_user(), stream)

    def update_mod_flag(self):
        old_list = self.mod_flag
        new_list = set([self.mid])

        if old_list is None:
            old_list = []

        # Collect all dependency IDs
        for pkg in self.packages:
            for dep in pkg.dependencies:
                mod = self._repo.query(dep['id'])
                if mod and mod.mtype not in ('tool', 'engine'):
                    new_list.add(dep['id'])

        # Remove old IDs which have been removed from our dependencies
        for i, mid in reversed(list(enumerate(old_list))):
            if mid not in new_list:
                del old_list[i]

        # Add new IDs which have been added to our dependencies
        for mid in new_list:
            if mid not in old_list:
                old_list.append(mid)

        self.mod_flag = old_list

    def get_mod_flag(self):
        # Since mod_flag is just a list of IDs, we have to look up their paths here.
        paths = []
        dev_involved = False
        mods = {}

        # We have to retrieve the mods this way to honor the version constraints
        for pkg in self._repo.process_pkg_selection(self.packages):
            mod = pkg.get_mod()

            if mod.mid not in mods:
                mods[mod.mid] = mod

        for mid in self.mod_flag:
            if mid not in mods:
                # We don't know if this is an optional dependency; ignore it for now.
                logging.debug('Skipping mod "%s" during -mod generation because it\'s missing.' % mid)
                continue

            mod = mods[mid]
            if mod.dev_mode:
                for pkg in mod.packages:
                    if pkg.check_env():
                        paths.append((os.path.join(mod.folder, pkg.folder), '%s - %s' % (mod.title, pkg.name)))
                        dev_involved = True

            else:
                paths.append((mod.folder, mod.title))

        return paths, dev_involved

    def get_executables(self, user=False):
        exes = []
        if user and self.user_custom_build:
                exes.append({
                    'file': self.user_custom_build,
                    'mod': self,
                    'label': None
                })

        if self.custom_build:
            exes.append({
                'file': self.custom_build,
                'mod': self,
                'label': None
            })

        if self.mtype in ('engine', 'tool'):
            deps = self.packages
        else:
            try:
                deps = self.resolve_deps(True, False)
            except ModNotFound:
                logging.exception('Error during dep resolve for executables!')
                deps = []

        if user and self.user_exe:
            try:
                exe_mod = self._repo.query(self.user_exe[0], util.Spec('==' + self.user_exe[1]))
                deps = exe_mod.packages + list(deps)
            except ModNotFound:
                pass

        for pkg in deps:
            mod = pkg.get_mod()
            if mod.mtype != 'engine' or (mod.dev_mode and not pkg.check_env()):
                continue

            for exe in pkg.executables:
                exe = exe.copy()
                pkgpath = mod.folder
                if mod.dev_mode:
                    pkgpath = os.path.join(pkgpath, pkg.folder)

                exe['file'] = os.path.join(pkgpath, exe['file'])
                exe['mod'] = mod
                exes.append(exe)

        if not exes:
            raise NoExecutablesFound('No engine found for "%s"!' % self.title)

        return exes


class IniMod(InstalledMod):
    _pr_list = None
    _sc_list = None

    def __init__(self, values=None):
        super(IniMod, self).__init__(values)

        self._pr_list = []
        self._sc_list = []

        self.version = semantic_version.Version('1.0.0')

    def load(self, path):
        with open(path, 'r') as stream:
            for line in stream:
                if '=' not in line:
                    continue

                line = line.split('=')
                name = line[0].strip()
                value = line[1].strip(' \r\n\t;')

                # Skip empty values
                if value == '':
                    continue

                if name == 'modname':
                    self.title = value
                elif name == 'infotext':
                    self.description = value
                elif name.startswith('image'):
                    self.logo = value
                elif name == 'primarylist':
                    self._pr_list = value.split(',')
                elif name in ('secondarylist', 'secondrylist'):
                    self._sc_list = value.split(',')

        self.folder = os.path.dirname(path)
        if self.title == '':
            self.title = os.path.basename(self.folder)

        self.mid = os.path.basename(self.folder)
        if self.logo:
            self.logo_path = os.path.join(path, self.logo)

        pkg = InstalledPackage({
            'name': 'Content',
            'status': 'required'
        }, self)

        # Set some default values
        self.screenshots = []
        self.attachments = []

        self.add_pkg(pkg)

    def get_mod_flag(self):
        mods = self._pr_list[:]
        mods.append(self.folder)
        mods.extend(self._sc_list)

        return mods

    def get_primary_list(self):
        return self._pr_list

    def get_secondary_list(self):
        return self._sc_list


class InstalledPackage(Package):
    check_notes = ''
    files_ok = -1
    files_checked = -1

    @staticmethod
    def convert(pkg, mod):
        return InstalledPackage(pkg.get(), mod)

    def set(self, values):
        super(InstalledPackage, self).set(values.copy())

        self.check_notes = values.get('check_notes', '')

    def get(self):
        data = super(InstalledPackage, self).get()
        data['check_notes'] = self.check_notes
        return data



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import os
import sys
import logging
import functools
import json

from . import uhf
uhf(__name__)

from . import center, util, integration, web, repo, ipc
from .qt import QtCore, QtWidgets, load_styles
from .ui.hell import Ui_MainWindow as Ui_Hell
from .ui.install import Ui_InstallDialog
from .ui.edit_description import Ui_Dialog as Ui_DescEditor
from .tasks import run_task, InstallTask, UninstallTask, WindowsUpdateTask, LoadLocalModsTask

# Keep references to all open windows to prevent the GC from deleting them.
_open_wins = []
translate = QtCore.QCoreApplication.translate


class QDialog(QtWidgets.QDialog):

    def __init__(self, *args):
        super(QDialog, self).__init__(*args)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)


class QMainWindow(QtWidgets.QMainWindow):

    def __init__(self, *args):
        super(QMainWindow, self).__init__(*args)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

    def closeEvent(self, e):
        if center.pmaster.is_busy():
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                'Some tasks are still running in the background. Please wait for them to finish or abort them.')

            e.ignore()
        else:
            e.accept()
            center.app.quit()

    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.ActivationChange:
            if integration.current is not None:
                integration.current.annoy_user(False)

            if center.main_win and center.main_win.win.isActiveWindow():
                center.auto_fetcher.trigger()

        return super(QMainWindow, self).changeEvent(event)


class Window(object):
    win = None
    closed = True
    _is_window = True
    _tpl_cache = None
    _styles = ''
    _qchildren = None

    def __init__(self, window=True):
        super(Window, self).__init__()

        self._is_window = window
        self._tpl_cache = {}
        self._qchildren = []

    def _create_win(self, ui_class, qt_widget=QDialog):
        if not self._is_window:
            qt_widget = QtWidgets.QWidget

        self.win = util.init_ui(ui_class(), qt_widget())

    def open(self):
        global _open_wins

        if not self.closed:
            return

        self.closed = False
        _open_wins.append(self)
        self.win.destroyed.connect(self._del)
        self.win.show()

    def close(self):
        self.win.close()

    def _del(self):
        global _open_wins

        if self.closed:
            return

        self.closed = True

        if self in _open_wins:
            _open_wins.remove(self)

    def label_tpl(self, label, **vars):
        name = label.objectName()
        if name in self._tpl_cache:
            text = self._tpl_cache[name]
        else:
            text = self._tpl_cache[name] = label.text()

        for name, value in vars.items():
            text = text.replace('{' + name + '}', value)

        label.setText(text)

    def load_styles(self, path, append=True):
        data = load_styles(path)

        if append:
            self._styles += data
        else:
            self._styles = data

        self.win.setStyleSheet(self._styles)

    def tr(self, *args):
        return translate(self.__class__.__name__, *args)


class HellWindow(Window):
    _tasks = None
    _mod_filter = 'explore'
    _search_text = ''
    _updating_mods = None
    _init_done = False
    _prg_visible = False
    browser_ctrl = None
    progress_win = None

    def __init__(self, window=True):
        super(HellWindow, self).__init__(window)
        self._tasks = {}
        self._updating_mods = {}

        self._create_win(Ui_Hell, QMainWindow)
        self.browser_ctrl = web.BrowserCtrl(self.win.webView)
        self.win.verticalLayout.setContentsMargins(0, 0, 0, 0)

        self.win.webView.loadStarted.connect(self.show_indicator)
        self.win.webView.loadFinished.connect(self.check_loaded)

        center.signals.update_avail.connect(self.ask_update)
        center.signals.task_launched.connect(self.watch_task)
        center.signals.repo_updated.connect(self.update_mod_list)

        self.win.setWindowTitle(self.win.windowTitle() + ' ' + center.VERSION)
        self.win.progressInfo.hide()
        self.open()

    def _del(self):
        center.signals.update_avail.disconnect(self.ask_update)
        center.signals.task_launched.disconnect(self.watch_task)

        # Trying to free this object usually leads to memory corruption
        # See http://pyqt.sourceforge.net/Docs/PyQt5/gotchas.html#crashes-on-exit
        # Since this is the main window and should only be closed whenever the application exits, skipping this
        # won't lead to memory leaks.
        # super(HellWindow, self)._del()

    def finish_init(self):
        if self._init_done:
            return

        self._init_done = True

        center.auto_fetcher.start()
        ipc.setup()

        run_task(LoadLocalModsTask())

    def ask_update(self, version):
        # We only have an updater for windows.
        if sys.platform.startswith('win'):
            msg = self.tr('There\'s an update available!\nDo you want to install Knossos %s now?') % str(version)
            buttons = QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
            result = QtWidgets.QMessageBox.question(center.app.activeWindow(), 'Knossos', msg, buttons)
            if result == QtWidgets.QMessageBox.Yes:
                run_task(WindowsUpdateTask())
        else:
            msg = self.tr('There\'s an update available!\nYou should update to Knossos %s.') % str(version)
            QtWidgets.QMessageBox.information(center.app.activeWindow(), 'Knossos', msg)

    def search_mods(self):
        mods = None

        if self._mod_filter in ('home', 'develop'):
            mods = center.installed.mods
        elif self._mod_filter == 'explore':
            mods = center.mods.mods
        else:
            mods = {}

        # Now filter the mods.
        query = self._search_text
        result = []
        for mid, mvs in mods.items():
            if query in mvs[0].title.lower():
                mod = mvs[0]
                if mod.mtype == 'engine' and self._mod_filter != 'develop':
                    mvs = [mv for mv in mvs if mv.stability == center.settings['engine_stability']]
                    if len(mvs) == 0:
                        mvs = mods[mid]

                    mod = mvs[0]

                item = mod.get()
                item['progress'] = 0
                item['progress_info'] = {}

                installed_versions = {}
                for m in center.installed.mods.get(mid, []):
                    installed_versions[str(m.version)] = m.dev_mode

                rmod = center.mods.mods.get(mid, [])
                if mod.mtype == 'engine':
                    rm_sel = None
                    for m in rmod:
                        if m.stability == center.settings['engine_stability']:
                            rm_sel = m
                            break

                    if rm_sel:
                        rmod = rm_sel
                    elif len(rmod) > 0:
                        rmod = rmod[0]
                elif len(rmod) > 0:
                    rmod = rmod[0]

                # TODO: Refactor (see also templates/kn-{details,devel}-page.vue)
                if rmod and rmod.version > mod.version:
                    item['status'] = 'update'
                elif mod.mid in self._updating_mods:
                    item['status'] = 'updating'
                    item['progress'] = self._updating_mods[mod.mid]
                else:
                    item['status'] = 'ready'

                    if self._mod_filter == 'home':
                        for pkg in mod.packages:
                            if pkg.files_checked > 0 and pkg.files_ok < pkg.files_checked:
                                item['status'] = 'error'
                                break

                item['installed'] = len(installed_versions) > 0
                item['versions'] = []
                for mod in mvs:
                    mv = mod.get()
                    mv['installed'] = mv['version'] in installed_versions
                    mv['dev_mode'] = installed_versions.get(mv['version'], False)
                    item['versions'].append(mv)

                result.append(item)

        result.sort(key=lambda m: m['title'])
        return result, self._mod_filter

    def update_mod_list(self):
        if center.settings['base_path'] is not None:
            result, filter_ = self.search_mods()

            if filter_ in ('home', 'explore', 'develop'):
                self.browser_ctrl.bridge.updateModlist.emit(json.dumps(result), filter_)

    def show_indicator(self):
        self.win.setCursor(QtCore.Qt.BusyCursor)

    def check_loaded(self, success):
        self.win.unsetCursor()

    def update_mod_buttons(self, clicked=None):
        if center.settings['base_path'] is not None:
            self._mod_filter = clicked
            self.update_mod_list()

    def perform_search(self, term):
        self._search_text = term.lower()
        self.update_mod_list()

    def get_tasks(self):
        return self._tasks

    def watch_task(self, task):
        logging.debug('Task "%s" (%d, %s) started.', task.title, id(task), task.__class__)
        self._tasks[id(task)] = task
        self.browser_ctrl.bridge.taskStarted.emit(id(task), task.title, [m.mid for m in task.mods])

        for m in task.mods:
            self._updating_mods[m.mid] = 0

        task.done.connect(functools.partial(self._forget_task, task))
        task.progress.connect(functools.partial(self._track_progress, task))

        if len(task.mods) == 0 and not self._prg_visible:
                self._prg_visible = True
                self.win.progressInfo.show()
                self.win.progressLabel.setText(task.title)
                self.win.progressBar.setValue(0)

                integration.current.show_progress(0)

    def _track_progress(self, task, pi):
        self.browser_ctrl.bridge.taskProgress.emit(id(task), pi[0] * 100, json.dumps(pi[1]))

        for m in task.mods:
            self._updating_mods[m.mid] = pi[0] * 100

        if len(task.mods) == 0 and self._prg_visible:
            integration.current.set_progress(pi[0])
            self.win.progressBar.setValue(pi[0] * 100)

    def _forget_task(self, task):
        logging.debug('Task "%s" (%d) finished.', task.title, id(task))
        self.browser_ctrl.bridge.taskFinished.emit(id(task))
        del self._tasks[id(task)]

        for m in task.mods:
            if m.mid in self._updating_mods:
                del self._updating_mods[m.mid]

        if len(task.mods) == 0 and self._prg_visible:
            self._prg_visible = False
            self.win.progressInfo.hide()
            integration.current.hide_progress()

    def abort_task(self, task):
        if task in self._tasks:
            self._tasks[task].abort()


class ModInstallWindow(Window):
    _mod = None
    _pkg_checks = None
    _dep_tpl = None

    def __init__(self, mod, sel_pkgs=[]):
        super(ModInstallWindow, self).__init__()

        self._mod = mod
        self._create_win(Ui_InstallDialog)
        self.win.splitter.setStretchFactor(0, 2)
        self.win.splitter.setStretchFactor(1, 1)

        self.label_tpl(self.win.titleLabel, MOD=mod.title)
        self.win.notesField.setPlainText(mod.notes)

        self._pkg_checks = []
        self.show_packages()
        self.update_deps()

        self.win.treeWidget.itemChanged.connect(self.update_selection)
        self.win.treeWidget.itemClicked.connect(self.update_notes)

        self.win.accepted.connect(self.install)
        self.win.rejected.connect(self.close)

        self.open()

    def show_packages(self):
        pkgs = self._mod.packages
        try:
            all_pkgs = center.mods.process_pkg_selection(pkgs)
        except repo.ModNotFound as exc:
            logging.exception('Well, I won\'t be installing that...')
            msg = self.tr('I\'m sorry but you won\'t be able to install "%s" because "%s" is missing!') % (
                self._mod.title, exc.mid)
            QtWidgets.QMessageBox.critical(None, 'Knossos', msg)

            self.close()
            return

        needed_pkgs = self._mod.resolve_deps()
        mods = set()

        for pkg in all_pkgs:
            mods.add(pkg.get_mod())

        # Make sure our current mod comes first.
        if self._mod in mods:
            mods.remove(self._mod)

        mods = [self._mod] + list(mods)
        for mod in mods:
            item = QtWidgets.QTreeWidgetItem(self.win.treeWidget, ['%s %s' % (mod.title, mod.version), ''])
            item.setExpanded(True)
            item.setData(0, QtCore.Qt.UserRole, mod)
            item.setData(0, QtCore.Qt.UserRole + 2, '')
            c = 0

            try:
                local_mod = center.installed.query(mod)
            except repo.ModNotFound:
                local_mod = None

            for pkg in mod.packages:
                fsize = 0
                for f_item in pkg.files.values():
                    fsize += f_item.get('filesize', 0)

                if fsize > 0:
                    fsize = util.format_bytes(fsize)
                else:
                    fsize = '?'

                sub = QtWidgets.QTreeWidgetItem(item, [pkg.name, fsize])
                is_installed = center.installed.is_installed(pkg)
                is_required = len(center.installed.get_dependents([pkg])) > 0

                if is_installed:
                    sub.setText(1, self.tr('Installed'))

                if pkg.status == 'required' or pkg in needed_pkgs or is_required or (local_mod and local_mod.dev_mode):
                    sub.setCheckState(0, QtCore.Qt.Checked)
                    sub.setDisabled(True)

                    c += 1
                elif pkg.status == 'recommended':
                    if not local_mod or is_installed:
                        sub.setCheckState(0, QtCore.Qt.Checked)
                        c += 1
                    else:
                        sub.setCheckState(0, QtCore.Qt.Unchecked)

                elif pkg.status == 'optional':
                    if is_installed:
                        sub.setCheckState(0, QtCore.Qt.Checked)
                        c += 1
                    else:
                        sub.setCheckState(0, QtCore.Qt.Unchecked)

                sub.setData(0, QtCore.Qt.UserRole, pkg)
                sub.setData(0, QtCore.Qt.UserRole + 1, is_installed)
                sub.setData(0, QtCore.Qt.UserRole + 2, pkg.notes)
                self._pkg_checks.append(sub)

            if c == len(mod.packages):
                item.setCheckState(0, QtCore.Qt.Checked)
            elif c > 0:
                item.setCheckState(0, QtCore.Qt.PartiallyChecked)
            else:
                item.setCheckState(0, QtCore.Qt.Unchecked)

        self.win.treeWidget.header().resizeSections(QtWidgets.QHeaderView.ResizeToContents)

    def update_selection(self, item, col):
        obj = item.data(0, QtCore.Qt.UserRole)
        if isinstance(obj, repo.Mod):
            cs = item.checkState(0)

            for i in range(item.childCount()):
                child = item.child(i)
                if not child.isDisabled():
                    child.setCheckState(0, cs)

        self.update_deps()

    def get_selected_pkgs(self):
        to_install = []
        to_remove = []

        for item in self._pkg_checks:
            selected = item.checkState(0) == QtCore.Qt.Checked
            installed = item.data(0, QtCore.Qt.UserRole + 1)
            name = item.data(0, QtCore.Qt.UserRole)

            if selected != installed:
                if installed:
                    to_remove.append(name)
                else:
                    to_install.append(name)

        to_install = center.mods.process_pkg_selection(to_install)
        return to_install, to_remove

    def update_deps(self):
        pkg_sel, rem_sel = self.get_selected_pkgs()
        dl_size = 0
        for item in self._pkg_checks:
            pkg = item.data(0, QtCore.Qt.UserRole)

            if not item.data(0, QtCore.Qt.UserRole + 1):
                if item.checkState(0) == QtCore.Qt.Checked or pkg in pkg_sel:
                    for f_item in pkg.files.values():
                        dl_size += f_item.get('filesize', 0)

            if item.isDisabled():
                continue

            if pkg in pkg_sel:
                item.setCheckState(0, QtCore.Qt.Checked)

        self.label_tpl(self.win.dlSizeLabel, DL_SIZE=util.format_bytes(dl_size))

    def update_notes(self, item):
        self.win.notesField.setPlainText(item.data(0, QtCore.Qt.UserRole + 2))

    def install(self):
        to_install, to_remove = self.get_selected_pkgs()

        run_task(InstallTask(to_install, self._mod))
        run_task(UninstallTask(to_remove, mods=[self._mod]))
        self.close()


class DescriptionEditorWindow(Window):

    def __init__(self, text):
        super(DescriptionEditorWindow, self).__init__()

        self._create_win(Ui_DescEditor)
        self.win.descEdit.setPlainText(text)

        self.win.applyButton.clicked.connect(self.apply_text)
        self.win.cancelButton.clicked.connect(self.close)

        self.open()

    def apply_text(self):
        center.main_win.browser_ctrl.bridge.applyDevDesc.emit(self.win.descEdit.toPlainText())
        self.close()



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import os
import sys
import logging
import subprocess
import shutil
import glob
import stat
import json
import tempfile
import threading
import random
import time
import re
import hashlib
import semantic_version

from . import center, util, progress, nebula, repo, vplib
from .repo import Repo
from .qt import QtCore, QtWidgets, read_file


translate = QtCore.QCoreApplication.translate


class FetchTask(progress.Task):

    def __init__(self):
        super(FetchTask, self).__init__()
        self.title = 'Fetching mod list...'
        self.done.connect(self.finish)
        self.add_work([('repo', i * 100, link[0]) for i, link in enumerate(center.settings['repos'])])

    def work(self, params):
        if params[0] == 'repo':
            _, prio, link = params

            progress.update(0.1, 'Fetching "%s"...' % link)

            try:
                raw_data = util.get(link, raw=True)
                if not raw_data:
                    return

                data = Repo()
                data.is_link = True
                data.base = os.path.dirname(raw_data.url)
                data.parse(raw_data.text)
            except Exception:
                logging.exception('Failed to decode "%s"!', link)
                return

            wl = []
            for mid, mvs in data.mods.items():
                for mod in mvs:
                    wl.append(('mod', mod))

            self.add_work(wl)
            self.post((prio, data))

    def finish(self):
        if not self.aborted:
            modlist = center.mods = Repo()
            res = self.get_results()
            res.sort(key=lambda x: x[0])

            for part in res:
                modlist.merge(part[1])

            modlist.save_json(os.path.join(center.settings_path, 'mods.json'))
            center.save_settings()
            center.main_win.update_mod_list()


class LoadLocalModsTask(progress.Task):
    can_abort = False
    _steps = 2

    def __init__(self):
        super(LoadLocalModsTask, self).__init__(threads=3)

        self.done.connect(self.finish)
        self.title = 'Loading installed mods...'

        if center.settings['base_path'] is None:
            logging.warning('A LoadLocalModsTask was launched even though no base path was set!')
        else:
            center.installed.clear()
            self.add_work((center.settings['base_path'],))
            self.add_work(center.settings['base_dirs'])

    def work(self, path):
        mods = center.installed

        subs = []
        mod_file = None

        try:
            for base in os.listdir(path):
                sub = os.path.join(path, base)

                if os.path.isdir(sub) and not sub.endswith('.dis'):
                    subs.append(sub)
                elif base.lower() == 'mod.json':
                    mod_file = sub
        except FileNotFoundError:
            logging.warning('The directory "%s" does not exist anymore!' % path)

        if mod_file:
            try:
                mod = repo.InstalledMod.load(mod_file)
                mods.add_mod(mod)
            except Exception:
                logging.exception('Failed to parse "%s"!', sub)

        self.add_work(subs)

    def finish(self):
        center.main_win.update_mod_list()


class CheckFilesTask(progress.MultistepTask):
    can_abort = False
    _mod = None
    _check_results = None
    _steps = 2

    def __init__(self, pkgs):
        super(CheckFilesTask, self).__init__()

        self.title = 'Checking %d packages...' % len(pkgs)
        self.pkgs = pkgs

        self.done.connect(self.finish)
        self._threads = 1

    def init1(self):
        pkgs = []

        for pkg in self.pkgs:
            mod = pkg.get_mod()
            pkgs.append((mod.folder, pkg))

        self.add_work(pkgs)

    def work1(self, data):
        modpath, pkg = data
        pkg_files = pkg.filelist
        count = float(len(pkg_files))
        success = 0
        checked = 0

        summary = {
            'ok': [],
            'corrupt': [],
            'missing': []
        }

        for info in pkg_files:
            mypath = util.ipath(os.path.join(modpath, info['filename']))
            if os.path.isfile(mypath):
                progress.update(checked / count, 'Checking "%s"...' % (info['filename']))

                if util.check_hash(info['checksum'], mypath, False):
                    success += 1
                    summary['ok'].append(info['filename'])
                else:
                    summary['corrupt'].append(info['filename'])
            else:
                summary['missing'].append(info['filename'])

            checked += 1

        self.post((pkg, success, checked, summary))

    def init2(self):
        # Save the results from step 1.
        self._check_results = self.get_results()

        self.add_work(('',))

    def work2(self, d):
        fnames = set()
        modpaths = set()
        loose = []

        # Collect all filenames
        for pkg, s, c, m in self._check_results:
            modpath = pkg.get_mod().folder
            modpaths.add(modpath)

            # Ignore files are generated by Knossos. Only mod.json files that are in the location where we expect it to
            # be are ignored. All other mod.json files will be considered loose
            fnames.add(os.path.join(modpath, 'mod.json'))

            for info in pkg.filelist:
                # relative paths are valid here but we only want the filename
                fnames.add(os.path.normpath(os.path.join(modpath, info['filename'])))

        # Check for loose files.
        for modpath in modpaths:
            for path, dirs, files in os.walk(modpath):
                for item in files:
                    name = os.path.join(path, item)
                    if name not in fnames and not item.startswith(('__k_plibs', 'knossos.')):
                        loose.append(name)

        self._check_results.append((None, 0, 0, {'loose': loose}))
        self._results = self._check_results

    def finish(self):
        bad_packages = []
        for result in self._check_results:
            if result[0] is None:
                # This is the entry which contains the loose files
                continue

            if result[1] != result[2]:
                # If the number of checked files is different than the number of valid files then there is something
                # wrong with this package
                bad_packages.append(result[0])

        if len(bad_packages) > 0:
            msg = "An error was detected while validating the game file integrity. The following packages are invalid:"
            for pkg in bad_packages:
                msg += "\n  - Package %s of mod %s" % (pkg.name, pkg.get_mod().title)
            msg += "\n\nThese mods are invalid and need to be redownloaded before they can be played without errors."

            QtWidgets.QMessageBox.critical(None, 'Knossos', msg)



# TODO: Optimize, make sure all paths are relative (no mod should be able to install to C:\evil)
# TODO: Add error messages.
class InstallTask(progress.MultistepTask):
    _pkgs = None
    _pkg_names = None
    _mods = None
    _dls = None
    _copies = None
    _steps = 4
    _error = False
    _7z_lock = None
    check_after = True

    def __init__(self, pkgs, mod=None, check_after=True):
        super(InstallTask, self).__init__()

        self._mods = set()
        self._pkgs = []
        self._pkg_names = []
        self.check_after = check_after

        if sys.platform == 'win32':
            self._7z_lock = threading.Lock()

        if mod is not None:
            self.mods = [mod]

        for pkg in pkgs:
            try:
                pmod = center.installed.query(pkg.get_mod())
                if pmod.dev_mode:
                    # Don't modify mods which are in dev mode!
                    continue
            except repo.ModNotFound:
                pass

            ins_pkg = center.installed.add_pkg(pkg)
            pmod = ins_pkg.get_mod()
            self._pkgs.append(ins_pkg)
            self._mods.add(pmod)
            self._pkg_names.append((pmod.mid, ins_pkg.name))

            for item in ins_pkg.files.values():
                self._slot_prog[id(item)] = ('%s: %s' % (pmod.title, item['filename']), 0, 'Checking...')

        for m in self._mods:
            if m not in self.mods:
                self.mods.append(m)

        center.signals.repo_updated.emit()
        self.done.connect(self.finish)
        self.title = 'Installing mods...'

    def abort(self):
        super(InstallTask, self).abort()

        util.cancel_downloads()

    def finish(self):
        if self.aborted:
            if self._cur_step == 1:
                # Need to remove all those temporary directories.
                for ar in self.get_results():
                    try:
                        shutil.rmtree(ar['tpath'])
                    except Exception:
                        logging.exception('Failed to remove "%s"!' % ar['tpath'])
            else:
                QtWidgets.QMessageBox.critical(None, 'Knossos',
                    self.tr('The mod installation was aborted before it could finish. ' +
                        'You might have to uninstall the partially installed mod(s).'))
        elif self._error:
            msg = self.tr(
                'An error occured during the installation of a mod. It might be partially installed.\n' +
                'If you need more help, ask ngld or read the debug log!'
            )
            QtWidgets.QMessageBox.critical(None, 'Knossos', msg)

        if not isinstance(self, UpdateTask):
            run_task(LoadLocalModsTask())

    def init1(self):
        if center.settings['neb_user']:
            try:
                neb = nebula.NebulaClient()
                editable = neb.get_editable_mods()

                for mod in self._mods:
                    if mod.mid in editable:
                        mod.dev_mode = True
            except Exception:
                logging.exception('Failed to login to the Nebula')

        self._threads = 3
        self.add_work(self._mods)

    def work1(self, mod):
        modpath = mod.folder
        mfiles = mod.get_files()
        mnames = [f['filename'] for f in mfiles] + ['knossos.bmp', 'mod.json']
        self._local.slot = id(mod)
        self._slot_prog[id(mod)] = (mod.title, 0, '')

        archives = set()
        progress.start_task(0, 0.9, '%s')
        progress.update(0, 'Checking %s...' % mod.title)

        kpath = os.path.join(modpath, 'mod.json')
        if os.path.isfile(kpath):
            try:
                with open(kpath, 'r') as stream:
                    info = json.load(stream)
            except Exception:
                logging.exception('Failed to parse mod.json!')
                info = None

            if info is not None and info['version'] != str(mod.version):
                logging.error('Overwriting "%s" (%s) with version %s.' % (mod.mid, info['version'], mod.version))

        if os.path.isdir(modpath):
            # TODO: Figure out if we want to handle these files (i.e. remove them)
            for path, dirs, files in os.walk(modpath):
                relpath = path[len(modpath):].lstrip('/\\')
                for item in files:
                    itempath = util.pjoin(relpath, item)
                    if not itempath.startswith('kn_') and itempath not in mnames:
                        logging.info('File "%s" is left over.', itempath)
        else:
            logging.debug('Folder %s for %s does not yet exist.', mod, modpath)
            os.makedirs(modpath)

        amount = float(len(mfiles))
        inst_mods = center.installed.query_all(mod.mid)
        copies = []
        pkg_folders = {}

        # query_all is a generator so the exception will be thrown when looping over the result
        try:
            for mv in inst_mods:
                if mv.dev_mode:
                    pf = pkg_folders.setdefault(mv, {})

                    for pkg in mv.packages:
                        pf[pkg.name] = pkg.folder
        except repo.ModNotFound:
            inst_mods = []

        for i, info in enumerate(mfiles):
            if (mod.mid, info['package']) not in self._pkg_names:
                continue

            progress.update(i / amount, 'Checking %s: %s...' % (mod.title, info['filename']))

            # Check if we already have this file
            found = False
            for mv in inst_mods:
                if mv.dev_mode:
                    itempath = util.ipath(os.path.join(mv.folder, pkg_folders[mv][info['package']], info['filename']))
                else:
                    itempath = util.ipath(os.path.join(mv.folder, info['filename']))

                if os.path.isfile(itempath) and util.check_hash(info['checksum'], itempath):
                    copies.append((mod, info['package'], info['filename'], itempath))
                    found = True
                    break

            if not found:
                archives.add((mod.mid, info['package'], info['archive']))
                logging.debug('%s: %s is missing/broken for %s.', info['package'], info['filename'], mod)

        self.post((archives, copies))
        progress.finish_task()
        progress.start_task(0.9, 0, 'Downloading logos...')

        # Make sure the images are in the mod folder so that they won't be deleted during the next
        # FetchTask.
        for prop in ('logo', 'tile', 'banner'):
            img_path = getattr(mod, prop)
            if img_path:
                ext = os.path.splitext(img_path)[1]
                dest = os.path.join(mod.folder, 'kn_' + prop + ext)

                if '://' in img_path:
                    # That's a URL
                    with open(dest, 'wb') as fobj:
                        util.download(img_path, fobj)

                setattr(mod, prop, dest)

        for prop in ('screenshots', 'attachments'):
            im_paths = getattr(mod, prop)
            for i, path in enumerate(im_paths):
                ext = os.path.splitext(path)[1]
                dest = os.path.join(mod.folder, 'kn_' + prop + '_' + str(i) + ext)

                if '://' in path:
                    with open(dest, 'wb') as fobj:
                        util.download(path, fobj)

                im_paths[i] = dest

        progress.finish_task()
        progress.update(1, 'Done')

    def init2(self):
        archives = set()
        copies = []
        downloads = []

        for a, c in self.get_results():
            archives |= a
            copies.extend(c)

        self._copies = copies

        for pkg in self._pkgs:
            mod = pkg.get_mod()
            for oitem in pkg.files.values():
                if (mod.mid, pkg.name, oitem['filename']) in archives:
                    item = oitem.copy()
                    item['mod'] = mod
                    item['pkg'] = pkg
                    item['_id'] = id(oitem)
                    downloads.append(item)
                else:
                    del self._slot_prog[id(oitem)]

        if len(archives) == 0:
            logging.info('Nothing to do for this InstallTask!')
        elif len(downloads) == 0:
            logging.error('Somehow we didn\'t find any downloads for this InstallTask!')
            self._error = True

        self._threads = 0
        self.add_work(downloads)

    def work2(self, archive):
        self._local.slot = archive['_id']

        with tempfile.TemporaryDirectory() as tpath:
            arpath = os.path.join(tpath, archive['filename'])
            modpath = archive['mod'].folder

            # TODO: Maybe this should be an option?
            retries = 3
            done = False
            urls = list(archive['urls'])
            random.shuffle(urls)

            while retries > 0:
                retries -= 1

                for url in urls:
                    progress.start_task(0, 0.97, '%s')
                    progress.update(0, 'Ready')

                    with open(arpath, 'wb') as stream:
                        if not util.download(url, stream):
                            if self.aborted:
                                return

                            logging.error('Download of "%s" failed!', url)
                            continue

                    progress.finish_task()
                    progress.update(0.97, 'Checking "%s"...' % archive['filename'])

                    if util.check_hash(archive['checksum'], arpath):
                        done = True
                        retries = 0
                        break
                    else:
                        logging.error('File "%s" is corrupted!', url)

            if not done:
                logging.error('Missing file "%s"!', archive['filename'])
                self._error = True
                return

            if self.aborted:
                return

            cpath = os.path.join(tpath, 'content')
            os.mkdir(cpath)

            needed_files = filter(lambda item: item['archive'] == archive['filename'], archive['pkg'].filelist)
            done = False

            if sys.platform == 'win32':
                # Apparently I can't run multiple 7z instances on Windows. If I do, I always get the error
                # "The archive can't be opened because it is still in use by another process."
                # I have no idea why. It works fine on Linux and Mac OS.
                # TODO: Is there a better solution?

                progress.update(0.98, 'Waiting...')
                self._7z_lock.acquire()

            progress.update(0.98, 'Extracting...')
            logging.debug('Extracting %s into %s', archive['filename'], modpath)

            if util.extract_archive(arpath, cpath):
                done = True
                # Look for missing files
                for item in needed_files:
                    src_path = os.path.join(cpath, item['orig_name'])

                    if not os.path.isfile(src_path):
                        logging.warning('Missing file "%s" from archive "%s" for package "%s" (%s)!',
                                        item['orig_name'], archive['filename'], archive['pkg'].name, archive['mod'].title)

                        done = False
                        break

            if sys.platform == 'win32':
                self._7z_lock.release()

            if not done:
                logging.error('Failed to unpack archive "%s" for package "%s" (%s)!',
                              archive['filename'], archive['pkg'].name, archive['mod'].title)
                shutil.rmtree(cpath, ignore_errors=True)
                self._error = True
                return

            dev_mode = archive['pkg'].get_mod().dev_mode

            for item in archive['pkg'].filelist:
                if item['archive'] != archive['filename']:
                    continue

                src_path = os.path.join(cpath, item['orig_name'])
                if dev_mode:
                    dest_path = util.ipath(os.path.join(modpath, archive['pkg'].folder, item['filename']))
                else:
                    dest_path = util.ipath(os.path.join(modpath, item['filename']))

                try:
                    dparent = os.path.dirname(dest_path)
                    if not os.path.isdir(dparent):
                        os.makedirs(dparent)

                    if dev_mode and archive['pkg'].is_vp:
                        progress.start_task(0.98, 0.02, '%s')
                        util.extract_vp_file(src_path, os.path.join(modpath, archive['pkg'].folder))

                        # Avoid confusing CheckTask with a missing VP file.
                        archive['pkg'].filelist = []
                    else:
                        # This move might fail on Windows with Permission Denied errors.
                        # "[WinError 32] The process cannot access the file because it is being used by another process"
                        # Just try it again several times to account of AV scanning and similar problems.
                        tries = 5
                        while tries > 0:
                            try:
                                shutil.move(src_path, dest_path)
                                break
                            except Exception as e:
                                logging.warning('Initial move for "%s" failed (%s)!' % (src_path, str(e)))
                                tries -= 1

                                if tries == 0:
                                    raise
                                else:
                                    time.sleep(1)
                except Exception:
                    logging.exception('Failed to move file "%s" from archive "%s" for package "%s" (%s) to its destination %s!',
                                      src_path, archive['filename'], archive['pkg'].name, archive['mod'].title, dest_path)
                    self._error = True

            # Copy the remaining empty dirs and symlinks.
            for path, dirs, files in os.walk(cpath):
                path = os.path.relpath(path, cpath)

                for name in dirs:
                    src_path = os.path.join(cpath, path, name)
                    dest_path = util.ipath(os.path.join(modpath, path, name))

                    if os.path.islink(src_path):
                        if not os.path.lexists(dest_path):
                            linkto = os.readlink(src_path)
                            os.symlink(linkto, dest_path)
                    elif not os.path.exists(dest_path):
                        os.makedirs(dest_path)

                for name in files:
                    src_path = os.path.join(cpath, path, name)

                    if os.path.islink(src_path):
                        dest_path = util.ipath(os.path.join(modpath, path, name))
                        if not os.path.lexists(dest_path):
                            linkto = os.readlink(src_path)
                            os.symlink(linkto, dest_path)

            progress.update(1, 'Done.')

    def init3(self):
        self.add_work((None,))

    def work3(self, _):
        self._slot_prog['copies'] = ('Copy old files', 0, 'Waiting...')
        self._local.slot = 'copies'

        pkg_folders = {}

        count = float(len(self._copies))
        for i, info in enumerate(self._copies):
            mod, pkg_name, fn, src = info

            progress.update(i / count, fn)
            if mod.dev_mode:
                if mod not in pkg_folders:
                    pkg_folders[mod] = {}
                    for pkg in mod.packages:
                        pkg_folders[mod][pkg.name] = pkg.folder

                dest = os.path.join(mod.folder, pkg_folders[mod][pkg_name], fn)
            else:
                dest = os.path.join(mod.folder, fn)

            logging.debug('Copying %s to %s', src, dest)
            shutil.copy(src, dest)

        progress.update(1, 'Done')

    def init4(self):
        self.add_work((None,))

    def work4(self, _):
        # Generate mod.json files.
        for mod in self._mods:
            try:
                mod.save()
                util.get(center.settings['nebula_link'] + 'api/1/track/install/' + mod.mid)
            except Exception:
                logging.exception('Failed to generate mod.json file for %s!' % mod.mid)


# TODO: make sure all paths are relative (no mod should be able to install to C:\evil)
class UninstallTask(progress.MultistepTask):
    _pkgs = None
    _mods = None
    _steps = 2
    check_after = True

    def __init__(self, pkgs, mods=[]):
        super(UninstallTask, self).__init__()

        self._pkgs = []
        self._mods = []

        for pkg in pkgs:
            try:
                self._pkgs.append(center.installed.query(pkg))
            except repo.ModNotFound:
                logging.exception('Someone tried to uninstall a non-existant package (%s, %s)! Skipping it...', pkg.get_mod().mid, pkg.name)

        for mod in mods:
            try:
                self._mods.append(center.installed.query(mod))
            except repo.ModNotFound:
                logging.exception('Someone tried to uninstall a non-existant %s!', mod)

        self.done.connect(self.finish)
        self.title = 'Uninstalling mods...'

    def init1(self):
        self.add_work(self._pkgs)

    def work1(self, pkg):
        mod = pkg.get_mod()

        for item in pkg.filelist:
            path = util.ipath(os.path.join(mod.folder, item['filename']))
            if not os.path.isfile(path):
                logging.warning('File "%s" for mod "%s" (%s) is missing during uninstall!', item['filename'], mod.title, mod.mid)
            else:
                os.unlink(path)

    def init2(self):
        mods = set(self._mods)

        # Unregister uninstalled pkgs.
        for pkg in self._pkgs:
            mods.add(pkg.get_mod())
            center.installed.del_pkg(pkg)

        self.add_work(mods)

    def work2(self, mod):
        modpath = mod.folder

        try:
            if isinstance(mod, repo.IniMod):
                shutil.rmtree(modpath)
            elif len(mod.packages) == 0:
                # Remove our files

                my_files = [os.path.join(modpath, 'mod.json'), mod.logo, mod.tile, mod.banner]
                my_files += mod.screenshots + mod.attachments
                for path in my_files:
                    if path and os.path.isfile(path):
                        os.unlink(path)

                libs = os.path.join(modpath, '__k_plibs')
                if os.path.isdir(libs):
                    # Delete any symlinks before running shutil.rmtree().
                    for link in os.listdir(libs):
                        item = os.path.join(libs, link)
                        if os.path.islink(item):
                            os.unlink(item)

                    shutil.rmtree(libs)

                center.installed.del_mod(mod)
            elif not os.path.isdir(modpath):
                logging.error('Mod %s still has packages but mod folder "%s" is gone!' % (mod, modpath))
            else:
                mod.save()
        except Exception:
            logging.exception('Failed to uninstall mod from "%s"!' % modpath)
            self._error = True

        # Remove empty directories.
        for path, dirs, files in os.walk(modpath, topdown=False):
            if len(dirs) == 0 and len(files) == 0:
                os.rmdir(path)

    def finish(self):
        # Update the local mod list which will remove the uninstalled mod
        run_task(LoadLocalModsTask())


class UpdateTask(InstallTask):
    _old_mod = None
    __check_after = True

    def __init__(self, mod, check_after=True):
        self._old_mod = mod
        new_mod = center.mods.query(mod.mid)
        self.__check_after = check_after

        old_pkgs = [pkg.name for pkg in mod.packages]
        pkgs = []

        for pkg in new_mod.packages:
            if pkg.name in old_pkgs or pkg.status == 'required':
                pkgs.append(pkg)

        super(UpdateTask, self).__init__(pkgs, new_mod, check_after=False)

    def finish(self):
        super(UpdateTask, self).finish()

        if not self.aborted and not self._error:
            # The new version has been succesfully installed, remove the old version.

            if len(self._old_mod.get_dependents()) == 0:
                run_task(UninstallTask(self._old_mod.packages))
            else:
                logging.debug('Not uninstalling %s after update because it still has dependents.', self._old_mod)
                run_task(LoadLocalModsTask())


class UploadTask(progress.MultistepTask):
    can_abort = True
    _steps = 2
    _client = None
    _mod = None
    _dir = None
    _login_failed = False
    _duplicate = False
    _reason = None
    _msg = None
    _success = False
    _msg_table = {
        'invalid version': 'the version number specified for this release is invalid!',
        'outdated version': 'there is already a release with the same or newer version on the nebula.',
        'unsupported archive checksum': 'your client sent an invalid checksum. You probably need to update.',
        'archive missing': 'one of your archives failed to upload.'
    }
    _question = QtCore.Signal()
    _question_result = False
    _question_cond = None

    def __init__(self, mod):
        super(UploadTask, self).__init__()

        self.title = 'Uploading mod...'
        self.mods = [mod]
        self._mod = mod.copy()

        self._threads = 2
        self._slot_prog = {
            'total': ('Status', 0, 'Waiting...')
        }

        self.done.connect(self.finish)
        self._question.connect(self.show_question)
        self._question_cond = threading.Condition()

    def abort(self):
        self._local.slot = 'total'
        progress.update(1, 'Aborted')

        super(UploadTask, self).abort()

    def show_question(self):
        res = QtWidgets.QMessageBox.question(None, 'Knossos', 'This mod has already been uploaded. If you continue, ' +
            'your metadata changes will be uploaded but the files will not be updated. Continue?')

        self._question_result = res == QtWidgets.QMessageBox.Yes
        with self._question_cond:
            self._question_cond.notify()

    def init1(self):
        self._local.slot = 'total'

        try:
            progress.update(0, 'Performing sanity checks...')
            if self._mod.mtype in ('mod', 'tc'):
                try:
                    exes = self._mod.get_executables()
                except Exception:
                    exes = []

                if len(exes) == 0:
                    self._reason = 'no exes'
                    self.abort()
                    return

            progress.update(0.1, 'Logging in...')

            self._dir = tempfile.TemporaryDirectory()
            self._client = client = nebula.NebulaClient()

            try:
                mods = client.get_editable_mods()
            except nebula.InvalidLoginException:
                self._login_failed = True
                self.abort()

                progress.update(0.1, 'Failed to login!')
                return

            progress.update(0.11, 'Updating metadata...')

            if self._mod.mid not in mods:
                client.create_mod(self._mod)
            else:
                client.update_mod(self._mod)

            progress.update(0.13, 'Performing pre-flight checks...')
            try:
                client.preflight_release(self._mod)
            except nebula.RequestFailedException as exc:
                if exc.args[0] == 'duplicated version':
                    with self._question_cond:
                        self._question.emit()
                        self._question_cond.wait()

                        if not self._question_result:
                            self._reason = 'aborted'
                            self.abort()
                            return

                    self._duplicate = True
                else:
                    raise

            if not self._duplicate:
                progress.update(0.15, 'Scanning files...')
                archives = []
                fnames = {}
                conflicts = {}
                for pkg in self._mod.packages:
                    ar_name = pkg.name + '.7z'
                    pkg_path = os.path.join(self._mod.folder, pkg.folder)
                    pkg.filelist = []

                    for sub, dirs, files in os.walk(pkg_path):
                        relsub = os.path.relpath(sub, pkg_path)
                        for fn in files:
                            relpath = os.path.join(relsub, fn).replace('\\', '/')

                            pkg.filelist.append({
                                'filename': relpath,
                                'archive': ar_name,
                                'orig_name': relpath,
                                'checksum': None
                            })

                            if not pkg.is_vp:
                                # VP conflicts don't cause problems and are most likely intentional
                                # which is why we ignore them.

                                if relpath in fnames:
                                    l = conflicts.setdefault(relpath, [fnames[relpath].name])
                                    l.append(pkg.name)

                                fnames[relpath] = pkg

                    if len(pkg.filelist) == 0:
                        self._reason = 'empty pkg'
                        self._msg = pkg.name
                        self.abort()
                        return

                    archives.append(pkg)
                    self._slot_prog[pkg.name] = (pkg.name + '.7z', 0, 'Waiting...')

                if conflicts:
                    msg = ''
                    for name in sorted(conflicts.keys()):
                        msg += '\n%s is in %s' % (name, util.human_list(conflicts[name]))

                    self._reason = 'conflict'
                    self._msg = msg
                    self.abort()
                    return

                self._slot_prog['#checksums'] = ('Checksums', 0, '')
                self._local.slot = '#checksums'

                fc = float(sum([len(pkg.filelist) for pkg in self._mod.packages]))
                done = 0
                for pkg in self._mod.packages:
                    pkg_path = os.path.join(self._mod.folder, pkg.folder)

                    for fn in pkg.filelist:
                        progress.update(done / fc, fn['filename'])
                        fn['checksum'] = util.gen_hash(os.path.join(pkg_path, fn['filename']))
                        done += 1

                progress.update(1, 'Done')
                self._local.slot = 'total'

                progress.update(0.2, 'Uploading...')
                self.add_work(archives)
        except nebula.AccessDeniedException:
            self._reason = 'unauthorized'
            self.abort()
        except nebula.RequestFailedException as exc:
            if exc.args[0] not in self._msg_table:
                logging.exception('Failed request to nebula during upload!')

            self._reason = exc.args[0]
            self.abort()
        except Exception:
            logging.exception('Error during upload initalisation!')
            self._reason = 'unknown'
            self.abort()

    def work1(self, pkg):
        self._local.slot = pkg.name

        ar_name = pkg.name + '.7z'
        ar_path = os.path.join(self._dir.name, ar_name)

        try:
            progress.update(0, 'Comparing...')

            hasher = hashlib.new('sha512')
            for item in sorted(pkg.filelist, key=lambda a: a['filename']):
                line = '%s#%s\n' % (item['filename'], item['checksum'])
                hasher.update(line.encode('utf8'))

            content_ck = hasher.hexdigest()
            result, meta = self._client.is_uploaded(content_checksum=content_ck)
            if result:
                # The file is already uploaded
                pkg.files[ar_name] = {
                    'filename': ar_name,
                    'dest': '',
                    'checksum': ('sha256', meta['checksum']),
                    'filesize': meta['filesize']
                }

                progress.update(1, 'Done!')
                return

            progress.update(0, 'Packing...')
            if pkg.is_vp:
                vp_name = pkg.name + '.vp'
                vp_path = os.path.join(self._dir.name, vp_name)
                vp = vplib.VpWriter(vp_path)
                pkg_path = os.path.join(self._mod.folder, pkg.folder)

                for sub, dirs, files in os.walk(pkg_path):
                    relsub = os.path.relpath(sub, pkg_path)
                    for fn in files:
                        relpath = os.path.join(relsub, fn)
                        vp.add_file(relpath, os.path.join(sub, fn))

                progress.start_task(0.0, 0.1, '%s')
                vp.write()

                progress.update(1, 'Calculating checksum...')
                progress.finish_task()

                pkg.filelist = [{
                    'filename': vp_name,
                    'archive': ar_name,
                    'orig_name': vp_name,
                    'checksum': util.gen_hash(vp_path)
                }]

                progress.start_task(0.1, 0.3, '%s')
            else:
                progress.start_task(0.0, 0.4, '%s')

            if pkg.is_vp:
                p = util.Popen([util.SEVEN_PATH, 'a', '-bsp1', ar_path, vp_name],
                    cwd=self._dir.name, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            else:
                p = util.Popen([util.SEVEN_PATH, 'a', '-bsp1', ar_path, '.'],
                    cwd=os.path.join(self._mod.folder, pkg.folder), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

            line_re = re.compile(r'^\s*([0-9]+)%')
            buf = ''
            while p.poll() is None:
                while '\b' not in buf:
                    line = p.stdout.read(10)
                    if not line:
                        break
                    buf += line.decode('utf8', 'replace')

                buf = buf.split('\b')
                line = buf.pop(0)
                buf = '\b'.join(buf)

                m = line_re.match(line)
                if m:
                    progress.update(int(m.group(1)) / 100., 'Compressing...')

            if p.returncode != 0:
                logging.error('Failed to build %s!' % ar_name)
                self.abort()
                return

            progress.finish_task()
            progress.start_task(0.4, 0.6, '%s')
            progress.update(0, 'Preparing upload...')

            pkg.files[ar_name] = {
                'filename': ar_name,
                'dest': '',
                'checksum': util.gen_hash(ar_path),
                'filesize': os.stat(ar_path).st_size
            }

            retries = 3
            while retries > 0:
                retries -= 1
                try:
                    self._client.upload_file(ar_name, ar_path, content_checksum=content_ck)
                    break
                except nebula.RequestFailedException:
                    logging.exception('Failed upload, retrying...')

            progress.finish_task()
            progress.update(1, 'Done!')
        except nebula.RequestFailedException:
            logging.exception('Failed request to nebula during upload!')

            self._reason = 'archive missing'
            self.abort()
        except Exception:
            logging.exception('Unknown error during package packing!')

            self._reason = 'unknown'
            self.abort()

    def init2(self):
        self._local.slot = 'total'

        try:
            progress.update(0.8, 'Finishing...')

            if self._duplicate:
                self._client.update_release(self._mod)
            else:
                self._client.create_release(self._mod)

            progress.update(1, 'Done')
            self._success = True
        except nebula.AccessDeniedException:
            self._reason = 'unauthorized'
            self.abort()

        except nebula.RequestFailedException as exc:
            if exc.args[0] not in self._msg_table:
                logging.exception('Failed request to nebula during upload!')

            self._reason = exc.args[0]
            self.abort()

    def work2(self):
        pass

    def finish(self):
        try:
            if self._dir:
                util.retry_helper(self._dir.cleanup)
        except OSError:
            # This is not a critical error so we only log it for now
            logging.exception('Failed to remove temporary folder after upload!')

        if self._login_failed:
            message = 'Failed to login!'
        elif self._reason == 'unauthorized':
            message = 'You are not authorized to edit this mod!'
        elif self._success:
            message = 'Successfully uploaded mod!'
        elif self._reason in self._msg_table:
            message = "Your mod couldn't be uploaded because %s" % self._msg_table[self._reason]
        elif self._reason == 'conflict':
            message = "I can't upload this mod because at least one file is contained in multiple packages.\n"
            message += self._msg
        elif self._reason == 'empty pkg':
            message = 'The package %s is empty!' % self._msg
        elif self._reason == 'no exes':
            message = 'The mod has no executables selected!'
        elif self._reason == 'aborted':
            return
        else:
            message = 'An unexpected error occured! Sorry...'

        center.main_win.browser_ctrl.bridge.taskMessage.emit(message)


class GOGExtractTask(progress.Task):
    can_abort = False

    def __init__(self, gog_path, dest_path):
        super(GOGExtractTask, self).__init__()

        self.done.connect(self.finish)
        self.add_work([(gog_path, dest_path)])
        self.title = 'Installing FS2 from GOG...'

        self._makedirs(dest_path)
        create_retail_mod(dest_path)
        center.main_win.update_mod_buttons('home')

        self.mods = [center.installed.query('FS2')]
        self._slot_prog['total'] = ('Status', 0, 'Waiting...')

        if not center.installed.has('FSO'):
            try:
                fso = center.mods.query('FSO')
                run_task(InstallTask(fso.resolve_deps()))
            except repo.ModNotFound:
                logging.warning('Installing retail files but FSO is missing!')

    def work(self, paths):
        gog_path, dest_path = paths
        self._local.slot = 'total'

        progress.update(0.03, 'Looking for InnoExtract...')
        data = util.get(center.INNOEXTRACT_LINK)

        try:
            data = json.loads(data)
        except Exception:
            logging.exception('Failed to read JSON data!')
            return

        link = None
        path = None
        for plat, info in data.items():
            if sys.platform.startswith(plat):
                link, path = info[:2]
                break

        if link is None:
            logging.error('Couldn\'t find an innoextract download for "%s"!', sys.platform)
            return

        if not os.path.exists(dest_path):
            os.makedirs(dest_path)

        inno = os.path.join(dest_path, os.path.basename(path))
        with tempfile.TemporaryDirectory() as tempdir:
            archive = os.path.join(tempdir, os.path.basename(link))

            progress.start_task(0.03, 0.10, 'Downloading InnoExtract...')
            with open(os.path.join(dest_path, archive), 'wb') as dl:
                util.download(link, dl)

            progress.finish_task()
            progress.update(0.13, 'Extracting InnoExtract...')

            util.extract_archive(archive, tempdir)
            shutil.move(os.path.join(tempdir, path), inno)

        # Make it executable
        mode = os.stat(inno).st_mode
        os.chmod(inno, mode | stat.S_IXUSR)

        progress.start_task(0.15, 0.75, 'Extracting FS2: %s')
        try:
            cmd = [inno, '-L', '-s', '-p', '-e', gog_path]
            logging.info('Running %s...', ' '.join(cmd))

            opts = dict()
            if sys.platform.startswith('win'):
                si = subprocess.STARTUPINFO()
                si.dwFlags = subprocess.STARTF_USESHOWWINDOW
                si.wShowWindow = subprocess.SW_HIDE

                opts['startupinfo'] = si
                opts['stdin'] = subprocess.PIPE

            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=dest_path, **opts)

            if sys.platform.startswith('win'):
                p.stdin.close()

            buf = ''
            while p.poll() is None:
                while '\r' not in buf:
                    line = p.stdout.read(10)
                    if not line:
                        break
                    buf += line.decode('utf8', 'replace')

                buf = buf.split('\r')
                line = buf.pop(0)
                buf = '\r'.join(buf)

                if 'MiB/s' in line:
                    try:
                        if ']' in line:
                            line = line.split(']')[1]

                        line = line.strip().split('MiB/s')[0] + 'MiB/s'
                        percent = float(line.split('%')[0]) / 100

                        progress.update(percent, line)
                    except Exception:
                        logging.exception('Failed to process InnoExtract output!')
                else:
                    if line.strip() == 'not a supported Inno Setup installer':
                        self.post(-1)
                        return

                    logging.info('InnoExtract: %s', line)
        except Exception:
            logging.exception('InnoExtract failed!')
            return

        progress.finish_task()

        progress.update(0.95, 'Moving files...')
        self._makedirs(os.path.join(dest_path, 'data/players'))
        self._makedirs(os.path.join(dest_path, 'data/movies'))

        for item in glob.glob(os.path.join(dest_path, 'app', '*.vp')):
            shutil.move(item, os.path.join(dest_path, os.path.basename(item)))

        for item in glob.glob(os.path.join(dest_path, 'app/data/players', '*.hcf')):
            shutil.move(item, os.path.join(dest_path, 'data/players', os.path.basename(item)))

        for item in glob.glob(os.path.join(dest_path, 'app/data2', '*.mve')):
            shutil.move(item, os.path.join(dest_path, 'data/movies', os.path.basename(item)))

        for item in glob.glob(os.path.join(dest_path, 'app/data3', '*.mve')):
            shutil.move(item, os.path.join(dest_path, 'data/movies', os.path.basename(item)))

        progress.update(0.99, 'Cleanup...')
        os.unlink(inno)
        shutil.rmtree(os.path.join(dest_path, 'app'), ignore_errors=True)
        shutil.rmtree(os.path.join(dest_path, 'tmp'), ignore_errors=True)

        self.post(dest_path)

    def _makedirs(self, path):
        if not os.path.isdir(path):
            os.makedirs(path)

    def finish(self):
        results = self.get_results()
        if len(results) < 1:
            QtWidgets.QMessageBox.critical(None, translate('tasks', 'Error'), self.tr(
                'The installer failed! Please read the log for more details...'))
            return
        elif results[0] == -1:
            QtWidgets.QMessageBox.critical(None, translate('tasks', 'Error'), self.tr(
                'The selected file wasn\'t a proper Inno Setup installer. Are you shure you selected the right file?'))
            return
        else:
            center.main_win.update_mod_list()
            center.main_win.browser_ctrl.bridge.retailInstalled.emit()


class GOGCopyTask(progress.Task):
    can_abort = False

    def __init__(self, gog_path, dest_path):
        super(GOGCopyTask, self).__init__()

        self.done.connect(self.finish)
        self.add_work([(gog_path, dest_path)])
        self.title = 'Copying retail files...'

        self._makedirs(dest_path)
        create_retail_mod(dest_path)
        center.main_win.update_mod_buttons('home')

        self.mods = [center.installed.query('FS2')]
        self._slot_prog['total'] = ('Status', 0, 'Waiting...')

        if not center.installed.has('FSO'):
            try:
                fso = center.mods.query('FSO')
                run_task(InstallTask(fso.resolve_deps()))
            except repo.ModNotFound:
                logging.warning('Installing retail files but FSO is missing!')

    def work(self, paths):
        gog_path, dest_path = paths
        self._local.slot = 'total'

        progress.update(0, 'Creating directories...')
        self._makedirs(os.path.join(dest_path, 'data/players'))
        self._makedirs(os.path.join(dest_path, 'data/movies'))

        progress.update(1 / 4., 'Copying VPs...')
        for item in glob.glob(os.path.join(gog_path, '*.vp')):
            shutil.copyfile(item, os.path.join(dest_path, os.path.basename(item)))

        progress.update(2 / 4., 'Copying player profiles...')
        for item in glob.glob(os.path.join(gog_path, 'data/players', '*.hcf')):
            shutil.copyfile(item, os.path.join(dest_path, 'data/players', os.path.basename(item)))

        progress.update(3 / 4., 'Copying cutscenes...')
        for item in glob.glob(os.path.join(gog_path, 'data2', '*.mve')):
            shutil.copyfile(item, os.path.join(dest_path, 'data/movies', os.path.basename(item)))

        for item in glob.glob(os.path.join(gog_path, 'data3', '*.mve')):
            shutil.copyfile(item, os.path.join(dest_path, 'data/movies', os.path.basename(item)))

        progress.update(1, 'Done')
        self.post(dest_path)

    def _makedirs(self, path):
        if not os.path.isdir(path):
            os.makedirs(path)

    def finish(self):
        center.main_win.update_mod_list()
        center.main_win.browser_ctrl.bridge.retailInstalled.emit()


class CheckUpdateTask(progress.Task):
    background = True

    def __init__(self):
        super(CheckUpdateTask, self).__init__()

        self.add_work(('',))
        self.title = 'Checking for updates...'

    def work(self, item):
        progress.update(0, 'Checking for updates...')

        update_base = util.pjoin(center.UPDATE_LINK, 'stable')
        version = util.get(update_base + '/version?me=' + center.VERSION)

        if version is None:
            logging.error('Update check failed!')
            return

        try:
            version = semantic_version.Version(version)
        except Exception:
            logging.exception('Failed to parse remote version!')
            return

        cur_version = semantic_version.Version(center.VERSION)
        if version > cur_version:
            center.signals.update_avail.emit(version)


class WindowsUpdateTask(progress.Task):

    def __init__(self):
        super(WindowsUpdateTask, self).__init__()

        self.done.connect(self.finish)
        self.add_work(('',))
        self.title = 'Installing update...'

    def work(self, item):
        # Download it.
        update_base = util.pjoin(center.UPDATE_LINK, 'stable')

        dir_name = tempfile.mkdtemp()
        updater = os.path.join(dir_name, 'knossos_updater.exe')

        progress.start_task(0, 0.98, 'Downloading update...')
        with open(updater, 'wb') as stream:
            util.download(update_base + '/updater.exe', stream)

        progress.finish_task()
        progress.update(0.99, 'Launching updater...')

        try:
            import win32api
            win32api.ShellExecute(0, 'open', updater, '/D=' + os.getcwd(), os.path.dirname(updater), 1)
        except Exception:
            logging.exception('Failed to launch updater!')
            self.post(False)
        else:
            self.post(True)
            center.app.quit()

    def finish(self):
        res = self.get_results()

        if len(res) < 1 or not res[0]:
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('Failed to launch the update!'))


class CopyFolderTask(progress.Task):

    def __init__(self, src_path, dest_path):
        super(CopyFolderTask, self).__init__()

        self.add_work(((src_path, dest_path),))
        self.title = 'Copying folder...'

    def work(self, p):
        src_path, dest_path = p

        if not os.path.isdir(src_path):
            logging.error('CopyFolderTask(): The src_path "%s" is not a folder!' % src_path)
            return

        progress.update(0, 'Scanning...')

        dest_base = os.path.dirname(dest_path)
        if not os.path.isdir(dest_base):
            os.makedirs(dest_base)

        plan = []
        total_size = 0.0
        for src_prefix, dirs, files in os.walk(src_path):
            dest_prefix = os.path.join(dest_path, os.path.relpath(src_prefix, src_path))

            for sub in dirs:
                sdest = os.path.join(dest_prefix, sub)
                try:
                    os.mkdir(sdest)
                except OSError:
                    logging.exception('Failed to mkdir %s.' % sdest)

            for sub in files:
                sdest = os.path.join(dest_prefix, sub)
                ssrc = os.path.join(src_prefix, sub)
                plan.append((ssrc, sdest))
                total_size += os.stat(ssrc).st_size

        bytes_done = 0
        for src, dest in plan:
            progress.update(bytes_done / total_size, os.path.relpath(src, src_path))
            shutil.copyfile(src, dest)

            bytes_done += os.stat(src).st_size


class VpExtractionTask(progress.Task):
    def __init__(self, installed_mod, ini_mod):
        super(VpExtractionTask, self).__init__()

        self.mod = installed_mod
        self.ini_mod = ini_mod

        self.title = 'Extracting VP files...'

        self._threads = 1  # VP extraction does not benefit from multiple threads

        for vp_file in os.listdir(ini_mod.folder):
            # We only look at vp files
            if not vp_file.lower().endswith(".vp"):
                continue

            vp_path = os.path.join(ini_mod.folder, vp_file)

            self.add_work((vp_path,))

    def work(self, vp_file):
        base_filename = os.path.basename(vp_file).replace(".vp", "")

        dest_folder = os.path.join(self.mod.folder, base_filename)

        progress.start_task(0.0, 1.0, 'Extracting %s')

        util.extract_vp_file(vp_file, dest_folder)

        progress.finish_task()
        # Collect the extracted vp files so we can use that once extraction has finished
        self.post(vp_file)


def run_task(task, cb=None):
    def wrapper():
        cb(task.get_results())

    if cb is not None:
        task.done.connect(wrapper)

    center.signals.task_launched.emit(task)
    center.pmaster.add_task(task)
    return task


def create_retail_mod(dest_path):
    # Remember to run tools/common/update_file_list.py if you add new files!
    files = {
        'tile': ':/html/images/retail_data/mod-retail.png',
        'banner': ':/html/images/retail_data/banner-retail.png',
    }

    screenshots = [':/html/images/retail_data/screen01.jpg', ':/html/images/retail_data/screen02.jpg', ':/html/images/retail_data/screen03.jpg', ':/html/images/retail_data/screen04.jpg', ':/html/images/retail_data/screen05.jpg', ':/html/images/retail_data/screen06.jpg', ':/html/images/retail_data/screen07.jpg', ':/html/images/retail_data/screen08.jpg', ':/html/images/retail_data/screen09.jpg', ':/html/images/retail_data/screen10.jpg', ':/html/images/retail_data/screen11.jpg', ':/html/images/retail_data/screen12.jpg']

    mod = repo.InstalledMod({
        'title': 'Retail FS2',
        'id': 'FS2',
        'version': '1.20',
        'type': 'tc',
        'description':
'[b][i]The year is 2367, thirty two years after the Great War. Or at least that is what YOU thought was the Great War. ' +
'The endless line of Shivan capital ships, bombers and fighters with super advanced technology was nearly overwhelming.\n\n' +
'As the Terran and Vasudan races finish rebuilding their decimated societies, a disturbance lurks in the not-so-far ' +
'reaches of the Gamma Draconis system.\n\nYour nemeses have arrived... and they are wondering what happened to ' +
'their scouting party.[/i][/b]\n\n[hr]FreeSpace 2 is a 1999 space combat simulation computer game developed by Volition as ' +
'the sequel to Descent: FreeSpace – The Great War. It was completed ahead of schedule in less than a year, and ' +
'released to very positive reviews.\n\nThe game continues on the story from Descent: FreeSpace, once again ' +
'thrusting the player into the role of a pilot fighting against the mysterious aliens, the Shivans. While defending ' +
'the human race and its alien Vasudan allies, the player also gets involved in putting down a rebellion. The game ' +
'features large numbers of fighters alongside gigantic capital ships in a battlefield fraught with beams, shells and ' +
'missiles in detailed star systems and nebulae.',
        'release_thread': 'http://www.hard-light.net/forums/index.php',
        'videos': ['https://www.youtube.com/watch?v=ufViyhrXzTE'],
        'first_release': '1999-09-30',
        'last_update': '1999-12-03',
        'folder': dest_path
    })

    mod.add_pkg(repo.InstalledPackage({
        'name': 'Content',
        'status': 'required',
        'folder': '.',
        'dependencies': [{
            'id': 'FSO',
            'version': '>=3.8.0-1'
        }]
    }))

    for prop, path in files.items():
        ext = os.path.splitext(path)[1]
        im_path = os.path.join(dest_path, 'kn_' + prop + ext)
        with open(im_path, 'wb') as stream:
            stream.write(read_file(path, decode=False))

        setattr(mod, prop, im_path)

    for i, path in enumerate(screenshots):
        ext = os.path.splitext(path)[1]
        im_path = os.path.join(dest_path, 'kn_screen_' + str(i) + ext)
        with open(im_path, 'wb') as stream:
            stream.write(read_file(path, decode=False))

        mod.screenshots.append(im_path)

    center.installed.add_mod(mod)
    mod.save()

    return mod



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import re
import threading
import subprocess
import time
import ctypes.util
import stat
import json
from subprocess import CalledProcessError

from . import center, repo, util, settings
from .qt import QtCore, QtWidgets, run_in_qt

# TODO: What happens if a SONAME contains a space?
FILE_PATH_RE = r'[a-zA-Z0-9\./\-\_\+]+'
LDD_RE = re.compile(r'\s*(' + FILE_PATH_RE + r') (?:=> (not found|' + FILE_PATH_RE + r'))?(?: \([^\)]+\))?\s*')
LIB_RE = re.compile(r'lib([a-zA-Z0-9\.\-\_\+]+)\.so(?:\..*)?')
LDCONF_RE = re.compile(r'\s*(' + FILE_PATH_RE + r') \([^\)]+\) => (' + FILE_PATH_RE + r')')
_LIB_CACHE = None

watchers = []
translate = QtCore.QCoreApplication.translate


class Fs2Watcher(threading.Thread):
    _params = None
    _cwd = None

    def __init__(self, params, cwd):
        super(Fs2Watcher, self).__init__()

        self._params = params
        self._cwd = cwd
        self.daemon = True
        self.start()

        watchers.append(self)

    def run(self):
        try:
            fs2_bin = os.path.normpath(self._params[0])

            if not os.path.isfile(fs2_bin):
                self.fs2_missing_msg(fs2_bin)
                return

            mode = os.stat(fs2_bin).st_mode
            if mode & stat.S_IXUSR != stat.S_IXUSR:
                # Make it executable.
                os.chmod(fs2_bin, mode | stat.S_IXUSR)

            env = os.environ.copy()
            if sys.platform.startswith('linux'):
                ld_path, missing = fix_missing_libs(fs2_bin)
                if len(missing) > 0:
                    self.complain_missing(missing)
                    return

                env['LD_LIBRARY_PATH'] = ld_path

            logging.debug('Launching FS2: %s in %s', repr([fs2_bin] + self._params[1:]), self._cwd)
            settings.ensure_fso_config()

            fail = False
            rc = -999
            reason = '???'
            try:
                p = subprocess.Popen([fs2_bin] + self._params[1:], cwd=self._cwd, env=env)

                time.sleep(0.3)
                if p.poll() is not None:
                    rc = p.returncode
                    reason = 'return code: %d' % rc
                    fail = True
            except OSError as exc:
                logging.exception('Failed to launch FS2!')
                reason = str(exc).decode('utf8', 'replace')
                fail = True

            if fail:
                center.signals.fs2_failed.emit(rc)
                self.failed_msg(reason, fs2_bin)
                return

            center.signals.fs2_launched.emit()
            p.wait()
            center.signals.fs2_quit.emit()
        finally:
            watchers.remove(self)

    @run_in_qt
    def failed_msg(self, reason, fs2_bin):
        msg = translate('runner', 'Starting %s failed! (%s)') % (fs2_bin, reason)
        QtWidgets.QMessageBox.critical(None, translate('runner', 'Failed'), msg)

    @run_in_qt
    def fs2_missing_msg(self, fs2_bin):
        QtWidgets.QMessageBox.critical(None, 'Knossos',
            translate('runner', 'I can\'t find the selected executable! (The file "%s" is missing!)') % fs2_bin)

    @run_in_qt
    def complain_missing(self, missing):
        if len(missing) > 1:
            msg = translate('runner', "I can't start because the libraries %s are missing!")
        else:
            msg = translate('runner', "I can't start because the library %s is missing!")

        QtWidgets.QMessageBox.critical(None, 'Knossos', msg % util.human_list(missing))


def check_elf_libs(fpath):
    out = util.check_output(['ldd', fpath], env={'LANG': 'C'}).splitlines()
    libs = {}

    for r_line in out:
        line = LDD_RE.match(r_line)
        if not line:
            logging.warning('Failed to parse ldd line "%s"!', r_line)
            continue

        if line.group(2) == 'not found':
            libs[line.group(1)] = None
        elif not line.group(2):
            # This is most likely a line like "linux-vdso.so.1 (0x00007fff847fe000)" which we can ignore.
            continue
        else:
            libs[line.group(1)] = line.group(2)

    return libs


def get_lib_path(filename):
    global _LIB_CACHE

    if not _LIB_CACHE:
        _LIB_CACHE = {}

        try:
            env = os.environ.copy()
            env['LANG'] = 'C'
            data = util.check_output(['ldconfig', '-p'], env=env).splitlines()
        except subprocess.CalledProcessError:
            logging.exception('Failed to run ldconfig!')
            return None

        for line in data[1:]:
            m = LDCONF_RE.match(line)
            if not m:
                logging.warning('Failed to parse ldconfig line "%s"!', line)
                continue

            _LIB_CACHE[m.group(1)] = m.group(2)

    return _LIB_CACHE.get(filename)


def fix_missing_libs(fpath, augment_ldpath=True):
    base = os.path.dirname(fpath)
    patch_dir = os.path.join(base, '__k_plibs')

    libs = check_elf_libs(fpath)
    missing = []
    for name, path in libs.items():
        if not path:
            missing.append(name)

    if len(missing) == 0:
        # Yay, nothing to do.
        if augment_ldpath:
            return os.environ.get('LD_LIBRARY_PATH', ''), missing
        else:
            return '', missing

    for lib in missing[:]:
        p_name = os.path.join(patch_dir, lib)
        if os.path.exists(p_name):
            missing.remove(lib)
        else:
            ld_name = LIB_RE.match(lib)
            if ld_name:
                fixed_name = get_lib_path(ctypes.util.find_library(ld_name.group(1)))
                if fixed_name:
                    if not os.path.isdir(patch_dir):
                        os.mkdir(patch_dir)

                    os.symlink(fixed_name, p_name)
                    missing.remove(lib)

    if augment_ldpath:
        ld_path = os.environ.get('LD_LIBRARY_PATH', '')
        if ld_path != '':
            patch_dir += ':' + ld_path

    return patch_dir, missing


def run_fs2_silent(params):
    base_path = center.settings['base_path']
    fs2_bin = params[0]

    if not os.path.isfile(fs2_bin):
        return -128, None

    mode = os.stat(fs2_bin).st_mode
    if mode & stat.S_IXUSR != stat.S_IXUSR:
        # Make it executable.
        os.chmod(fs2_bin, mode | stat.S_IXUSR)

    # We copy the existing environment to avoid omitting any values that may be necessary for running the binary
    # properly
    env = dict(os.environ)
    # AppImages are a bit special since they are themselves ELF executables which contain the actual FSO executable so
    # it makes no sense to examine the binary for missing libraries
    if sys.platform.startswith('linux') and not fs2_bin.lower().endswith(".appimage"):
        ld_path, missing = fix_missing_libs(fs2_bin)
        if len(missing) > 0:
            return -127, None

        env['LD_LIBRARY_PATH'] = ld_path

    try:
        try:
            output = util.check_output(params, env=env, cwd=base_path, stderr=subprocess.DEVNULL)
            rc = 0
        except CalledProcessError as e:
            # check_output raises this error if the return code was anything other than 0
            rc = e.returncode
            output = e.output
    except OSError:
        return -129, None

    if rc == 3221225595:
        # We're missing a DLL
        return -127, None

    return rc, output


def run_mod(mod, tool=None, exe_label=None):
    global installed

    if mod is None:
        mod = repo.Mod()

    try:
        inst_mod = center.installed.query(mod)
    except repo.ModNotFound:
        inst_mod = None

    if not inst_mod:
        QtWidgets.QMessageBox.critical(None, translate('runner', 'Error'),
            translate('runner', 'The mod "%s" could not be found!') % mod)
        return

    if tool:
        try:
            exes = tool.get_executables()
        except Exception:
            logging.exception('Failed to retrieve binaries for "%s"!' % tool.mid)
            QtWidgets.QMessageBox.critical(None, translate('runner', 'Error'),
                translate('runner', "I couldn't find the tool's executables! Aborted."))
            return
    else:
        try:
            exes = mod.get_executables(user=True)
        except Exception:
            logging.exception('Failed to retrieve binaries for "%s"!' % mod.mid)
            QtWidgets.QMessageBox.critical(None, translate('runner', 'Error'),
                translate('runner', "I couldn't find an executable. Aborted."))
            return

    exes = [item for item in exes if item.get('label', None) == exe_label]
    if len(exes) == 0:
        logging.error('"%s" provided no valid executable.' % (tool or mod).mid)
        QtWidgets.QMessageBox.critical(None, 'Knossos',
            translate('runner', 'No matching executable was found!'))
        return

    try:
        mod_flag, mod_choice = mod.get_mod_flag()
    except repo.ModNotFound:
        QtWidgets.QMessageBox.critical(None, 'Knossos',
            translate('runner', '"%s" has an error in its dependencies. Aborted.' % mod))
        return

    if mod_choice:
        # We have to ask the user
        center.main_win.browser_ctrl.bridge.showLaunchPopup.emit(json.dumps({
            'id': mod.mid,
            'version': str(mod.version),
            'title': mod.title,
            'exes': [(x['file'], '%s - %s' % (x['mod'].title, os.path.basename(x['file']))) for x in exes],
            'mod_flag': mod_flag
        }))
    else:
        run_mod_ex(mod, exes[0]['file'], [path for path, label in mod_flag])


def run_mod_ex(mod, binpath, mod_flag):
    # Put the cmdline together
    cmdline = mod.cmdline

    if mod.user_cmdline:
        cmdline = mod.user_cmdline

    if mod.mtype == 'mod':
        parent = mod.get_parent()

        if parent.mid == 'FS2':
            basepath = parent.folder
        else:
            basepath = os.path.dirname(parent.folder)

            found = False
            for pkg in mod.packages:
                for dep in pkg.dependencies:
                    if dep['id'] == parent.mid:
                        found = True
                        break

                if found:
                    break

            if not found:
                logging.debug("Mod %s doesn't depend on parent, fixing mod flags...", mod)
                mod_flag.extend([p[0] for p in parent.get_mod_flag()[0]])
    elif mod.mtype:
        basepath = mod.folder

    if len(mod_flag) > 0:
        # The paths for -mod must be relative to the base path.
        # TODO: Do we have to make sure that there are no special characters here or are the rules for mod.folder and
        #       pkg.folder enough to assure that?
        cmdline += ' -mod ' + ','.join([os.path.relpath(p, basepath) for p in mod_flag])

    # Look for the cmdline path.
    path = os.path.join(settings.get_fso_profile_path(), 'data/cmdline_fso.cfg')

    # Create the containing folders if they are missing.
    if not os.path.isfile(path):
        basep = os.path.dirname(path)
        if not os.path.isdir(basep):
            os.makedirs(basep)

    try:
        with open(path, 'w') as stream:
            stream.write(cmdline)
    except Exception:
        logging.exception('Failed to modify "%s". Not starting!!', path)

        QtWidgets.QMessageBox.critical(None, translate('runner', 'Error'),
            translate('runner', 'Failed to edit "%s"! I can\'t change the current mod!') % path)
    else:
        logging.info('Starting mod "%s" with cmdline "%s" and tool "%s".', mod.title, cmdline, binpath)
        Fs2Watcher([binpath], cwd=basepath)


def stringify_cmdline(line):
    result = []
    for part in line:
        if '"' in part:
            raise Exception("It's impossible to pass double quotes to FSO!")

        if ' ' in part:
            part = '"%s"' % part

        result.append(part)

    return ' '.join(result)



from ply import lex, yacc

tokens = ('LPARENS', 'RPARENS', 'AND', 'OR', 'NOT', 'VAR')

t_LPARENS = r'\('
t_RPARENS = r'\)'
t_AND = r'&&'
t_OR = r'\|\|'
t_NOT = r'!'
t_VAR = '[a-zA-Z_0-9]+'

t_ignore = ' \t\n\r'


def t_error(t):
    raise Exception('Invalid character %s!' % t.value[0])


precedence = (
    ('left', 'AND', 'OR'),
    ('right', 'NOT')
)


def p_expr_var(t):
    'expr : VAR'
    t[0] = ('ident', t[1])


def p_expr_parens(t):
    'expr : LPARENS expr RPARENS'
    t[0] = t[2]


def p_expr_not(t):
    'expr : NOT expr'
    t[0] = ('not', t[2])


def p_expr_op(t):
    """expr : expr AND expr
            | expr OR  expr"""

    t[0] = ('and' if t[2] == '&&' else 'or', t[1], t[3])


lexer = lex.lex()
parser = yacc.yacc()


def eval_expr(expr, values):
    if expr[0] == 'ident':
        return values.get(expr[1], False)
    elif expr[0] == 'not':
        return not eval_expr(expr[1], values)
    elif expr[0] == 'and':
        return eval_expr(expr[1], values) and eval_expr(expr[2], values)
    elif expr[0] == 'or':
        return eval_expr(expr[1], values) or eval_expr(expr[2], values)
    else:
        raise Exception('Invalid operation "%s" encountered!' % expr[0])


def eval_string(data, values):
    return eval_expr(parser.parse(data), values)


if __name__ == '__main__':
    import sys
    import json
    print(eval_string(sys.argv[1], json.loads(sys.argv[2])))



#!/usr/bin/python
## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function
import sys

if __package__ is None and not hasattr(sys, 'frozen'):
    import os.path
    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))


if len(sys.argv) > 1 and sys.argv[1] == '--cpuinfo':
    # We don't need to initialize knossos if we only need to fetch the CPU info.

    import json
    from knossos.third_party import cpuinfo

    info = None

    try:
        info = cpuinfo.get_cpu_info()
    except Exception:
        from knossos.launcher import logging
        logging.exception('Failed to retrieve CPU info.')

    print(json.dumps(info))
if len(sys.argv) > 1 and sys.argv[1] == '--run-cpuid':
    from knossos.third_party import cpuinfo

    print(cpuinfo.actual_get_cpu_info_from_cpuid())
elif len(sys.argv) > 1 and sys.argv[1] == '--deviceinfo':
    import json
    from knossos import clibs

    clibs.init_sdl()
    clibs.init_openal()

    if clibs.can_detect_audio():
        audio_devs = clibs.list_audio_devs()
    else:
        audio_devs = None

    print(json.dumps({
        'modes': clibs.get_modes(),
        'audio_devs': audio_devs,
        'joysticks': clibs.list_joysticks()
    }))
elif len(sys.argv) > 1 and sys.argv[1] == '--fso-config-path':
    from knossos import clibs

    clibs.init_sdl()
    print(clibs.get_config_path())
elif len(sys.argv) > 1 and sys.argv[1] == '--lib-paths':
    import json
    from knossos import clibs, center

    if len(sys.argv) > 3:
        if sys.argv[2] == 'auto':
            center.settings['sdl2_path'] = None
        else:
            center.settings['sdl2_path'] = sys.argv[2]

        if sys.argv[3] == 'auto':
            center.settings['openal_path'] = None
        else:
            center.settings['openal_path'] = sys.argv[3]

    try:
        clibs.init_sdl()
    except Exception:
        clibs.sdl = None

    try:
        clibs.init_openal()
    except Exception:
        clibs.acl = None

    if center.settings['sdl2_path'] and clibs.sdl:
        if clibs.sdl._name != center.settings['sdl2_path']:
            clibs.sdl = None

    if center.settings['openal_path'] and clibs.alc:
        if clibs.alc._name != center.settings['openal_path']:
            clibs.alc = None

    print(json.dumps({
        'sdl2': clibs.sdl._name if clibs.sdl else None,
        'openal': clibs.alc._name if clibs.alc else None
    }))
else:
    from knossos import launcher
    launcher.main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/hell.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1116, 800)
        MainWindow.setMinimumSize(QtCore.QSize(1116, 700))
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.webView = QtWebEngineWidgets.QWebEngineView(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.webView.sizePolicy().hasHeightForWidth())
        self.webView.setSizePolicy(sizePolicy)
        self.webView.setProperty("url", QtCore.QUrl("about:blank"))
        self.webView.setObjectName("webView")
        self.verticalLayout.addWidget(self.webView)
        self.progressInfo = QtWidgets.QWidget(self.centralwidget)
        self.progressInfo.setObjectName("progressInfo")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.progressInfo)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.progressLabel = QtWidgets.QLabel(self.progressInfo)
        self.progressLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.progressLabel.setObjectName("progressLabel")
        self.verticalLayout_2.addWidget(self.progressLabel)
        self.progressBar = QtWidgets.QProgressBar(self.progressInfo)
        self.progressBar.setProperty("value", 24)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_2.addWidget(self.progressBar)
        self.verticalLayout.addWidget(self.progressInfo)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1116, 25))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Knossos"))
        self.progressLabel.setText(_translate("MainWindow", "TextLabel"))

from ..qt import QtWebEngineWidgets




# parsetab.py
# This file is automatically generated. Do not edit.
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'leftANDORrightNOTLPARENS RPARENS AND OR NOT VARexpr : VARexpr : LPARENS expr RPARENSexpr : NOT exprexpr : expr AND expr\n            | expr OR  expr'
    
_lr_action_items = {'VAR':([0,3,4,5,6,],[2,2,2,2,2,]),'LPARENS':([0,3,4,5,6,],[3,3,3,3,3,]),'NOT':([0,3,4,5,6,],[4,4,4,4,4,]),'$end':([1,2,8,9,10,11,],[0,-1,-3,-4,-5,-2,]),'AND':([1,2,7,8,9,10,11,],[5,-1,5,-3,-4,-5,-2,]),'OR':([1,2,7,8,9,10,11,],[6,-1,6,-3,-4,-5,-2,]),'RPARENS':([2,7,8,9,10,11,],[-1,11,-3,-4,-5,-2,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expr':([0,3,4,5,6,],[1,7,8,9,10,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expr","S'",1,None,None,None),
  ('expr -> VAR','expr',1,'p_expr_var','bool_parser.py',26),
  ('expr -> LPARENS expr RPARENS','expr',3,'p_expr_parens','bool_parser.py',31),
  ('expr -> NOT expr','expr',2,'p_expr_not','bool_parser.py',36),
  ('expr -> expr AND expr','expr',3,'p_expr_op','bool_parser.py',41),
  ('expr -> expr OR expr','expr',3,'p_expr_op','bool_parser.py',42),
]



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import time
import logging

from threading import Thread, Condition
from knossos import center, tasks, qt


class AutoFetcher(Thread):
    _interval = 60 * 60  # = 1 hour
    _inactive_block = None

    def __init__(self):
        super(AutoFetcher, self).__init__()

        self._inactive_block = Condition()
        self.daemon = True

    def trigger(self):
        with self._inactive_block:
            self._inactive_block.notify()

    def run(self):
        while True:
            self.launch_task()
            time.sleep(self._interval)

            if not center.main_win.win.isActiveWindow():
                logging.debug('AutoFetcher paused.')
                with self._inactive_block:
                    self._inactive_block.wait()

                logging.debug('AutoFetcher resumed.')

    @qt.run_in_qt
    def launch_task(self):
        tasks.run_task(tasks.FetchTask())



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import print_function

import struct
import os
import logging
import shutil

from time import time

from knossos import progress
# See also: http://docs.python.org/library/struct.html#format-characters


class DataReader(object):
    _file = None
    _size = 0
    failed = False

    def __init__(self, filename):
        if isinstance(filename, str):
            self._file = open(filename, 'rb')
        else:
            self._file = filename

        self._file.seek(0, os.SEEK_END)
        self._size = self._file.tell()
        self._file.seek(0)

        self.read()

    def unpack(self, fmt):
        fmt = '<' + fmt
        try:
            size = struct.calcsize(fmt)
        except MemoryError:
            logging.error('Can\'t read data! Format: {0} [Memory]'.format(fmt))
            raise

        data = self._file.read(size)
        if len(data) < size:
            if hasattr(self._file, 'name'):
                name = self._file.name
            else:
                name = '<stream>'
            logging.error('File "{0}" is too short! Couldn\'t read "{1}". Position: {2}'.format(name, fmt, self._file.tell()))

        data = struct.unpack(fmt, data)

        if len(fmt) == 2:
            return data[0]
        else:
            return data

    def unpack_string(self):
        length = self.unpack('i')
        return self.unpack(str(length) + 's')[0].decode('utf8').rstrip('\x00')


class DataWriter(object):
    _file = None

    def __init__(self, filename):
        if isinstance(filename, str):
            self._file = open(filename, 'wb')
        else:
            self._file = filename

    def pack(self, fmt, *values):
        fmt = '<' + fmt
        self._file.write(struct.pack(fmt, *values))

    def pack_string(self, value):
        value += '\x00'
        length = len(value)

        self.pack('i', length)
        self.pack(str(length) + 's', value)


class VpReader(DataReader):
    fs = None
    files = None

    def cut_str(self, string):
        length = string.find(b'\x00')
        if length == -1:
            return string

        return string[:length]

    def read(self):
        header, version, diroffset, direntries = self.unpack('4siii')

        if header != b'VPVP':
            logging.error('Invalid VP file!')
            self._file.close()
            return

        dirs = 0
        files = 0

        self._file.seek(diroffset)
        self.files = {}
        cur_path = []
        for n in range(direntries):
            offset, size, name, timestamp = self.unpack('ii32si')
            name = self.cut_str(name).decode('utf8').lower()

            if timestamp == 0:
                if size != 0:
                    logging.error('Invalid directory entry "{0}"!'.format(name))
                    continue

                if name == '..':
                    cur_path.pop()
                else:
                    cur_path.append(name)
                    dirs += 1
            else:
                cur_path.append(name)
                self.files['/'.join(cur_path)] = {
                    'offset': offset,
                    'size': size,
                    'timestamp': timestamp
                }
                cur_path.pop()
                files += 1

        logging.info('Found {0} files and {1} directories in "{2}".'.format(files, dirs, self._file.name))

    def open_file(self, path):
        self._file.seek(self.files[path]['offset'])
        return self._file


class VpWriter(DataWriter):
    _files = None
    _count = 0
    _written = 0

    def __init__(self, path):
        super(VpWriter, self).__init__(path)

        self._files = dict()

    def add_file(self, path, content):
        path = path.replace('\\', '/').split('/')
        name = path.pop()
        lvl = self._files

        while len(path) > 0:
            lvl = lvl.setdefault(path.pop(0), {})

        lvl[name] = ('file', content)
        self._count += 1

    def get_file_count(self):
        return self._count

    def write_file(self, name, content, toc):
        progress.update(self._written / float(self._count), 'Packing "%s"...' % name)
        self._written += 1

        if isinstance(content, str):
            content = open(content, 'rb')
            opened = True
        else:
            opened = False

        offset = self._file.tell()
        shutil.copyfileobj(content, self._file)
        size = self._file.tell() - offset

        if opened:
            content.close()

        toc.append({
            'name': name,
            'offset': offset,
            'size': size,
            'timestamp': int(time())
        })

    def write_dir(self, lvl, toc):
        for name, content in lvl.items():
            if isinstance(content, tuple) and content[0] == 'file':
                # a file
                self.write_file(name, content[1], toc)
            else:
                toc.append({
                    'name': name,
                    'offset': 0,
                    'size': 0,
                    'timestamp': 0
                })
                self.write_dir(content, toc)
                toc.append({
                    'name': '..',
                    'offset': 0,
                    'size': 0,
                    'timestamp': 0
                })

    def write(self):
        self.pack('4siii', b'VPVP', 0, 0, 0)

        toc = []
        self.write_dir(self._files, toc)

        diroffset = self._file.tell()
        dirs = 0
        files = 0

        progress.update(0.99, 'Writing TOC...')
        for item in toc:
            self.pack('ii32si', item['offset'], item['size'], item['name'].encode('utf8'), item['timestamp'])

            if item['timestamp'] == 0:
                dirs += 1
            else:
                files += 1

        self._file.seek(0)
        self.pack('4siii', b'VPVP', 2, diroffset, len(toc))
        self._file.close()

        logging.info('Wrote {0} files and {1} directories in "{2}".'.format(files, dirs, self._file.name))



import os.path
import logging
import requests

from requests_toolbelt.multipart.encoder import MultipartEncoder, MultipartEncoderMonitor
from . import center, progress, util


class InvalidLoginException(Exception):
    pass


class RequestFailedException(Exception):
    pass


class AccessDeniedException(RequestFailedException):
    pass


class NebulaClient(object):
    _token = None
    _sess = None

    def __init__(self):
        self._sess = requests.Session()

    def _call(self, path, method='POST', skip_login=False, check_code=False, **kwargs):
        url = center.settings['nebula_link'] + path

        if not skip_login and not self._token:
            if not self.login():
                raise InvalidLoginException()

        if self._token:
            headers = kwargs.setdefault('headers', {})
            headers['X-KN-TOKEN'] = self._token

        try:
            result = self._sess.request(method, url, **kwargs)
        except requests.RequestException:
            logging.exception('Failed to send %s request to %s!' % (method, path))
            raise RequestFailedException('connection')

        if check_code and result.status_code != 200:
            reason = 'unknown'
            if result.status_code == 404:
                reason = 'not found'

            raise RequestFailedException(reason)

        return result

    def login(self, user=None, password=None):
        if not user:
            user = center.settings['neb_user']
            password = center.settings['neb_password']

        result = self._call('login', skip_login=True, data={
            'user': user,
            'password': password
        })

        if result.status_code != 200:
            return False

        data = result.json()
        if data['result']:
            self._token = data['token']
            return True
        else:
            return False

    def register(self, user, password, email):
        self._call('register', skip_login=True, check_code=True, data={
            'name': user,
            'password': password,
            'email': email
        })

        return True

    def reset_password(self, user):
        self._call('reset_password', skip_login=True, check_code=True, data={'user': user})
        return True

    def get_editable_mods(self):
        result = self._call('mod/editable', 'GET', check_code=True)
        return result.json()['mods']

    def _upload_mod_logos(self, mod):
        chks = [None, None]

        for i, prop in enumerate(('logo', 'tile')):
            im = getattr(mod, prop)
            if im and os.path.isfile(im):
                chks[i] = util.gen_hash(im)[1]
                self.upload_file(prop, im)

        return chks

    def check_mod_id(self, mid, title=None):
        result = self._call('mod/check_id', check_code=True, data={
            'id': mid,
            'title': title
        })

        return result.json()['result']

    def create_mod(self, mod):
        logo_chk, tile_chk = self._upload_mod_logos(mod)

        self._call('mod/create', check_code=True, json={
            'id': mod.mid,
            'title': mod.title,
            'type': mod.mtype,
            'parent': mod.parent,
            'logo': logo_chk,
            'tile': tile_chk,
            'first_release': mod.first_release.strftime('%Y-%m-%d') if mod.first_release else None,
            'members': []
        })
        return True

    def update_mod(self, mod):
        # TODO: Check if these actually changed
        logo_chk, tile_chk = self._upload_mod_logos(mod)

        self._call('mod/update', check_code=True, json={
            'id': mod.mid,
            'title': mod.title,
            'logo': logo_chk,
            'tile': tile_chk,
            'first_release': mod.first_release.strftime('%Y-%m-%d') if mod.first_release else None,
            'members': [center.settings['neb_user']]
        })
        return True

    def preflight_release(self, mod):
        meta = mod.get()
        meta['screenshots'] = []
        meta['attachments'] = []
        meta['banner'] = ''

        result = self._call('mod/release/preflight', check_code=True, json=meta)
        data = result.json()
        if not data:
            raise RequestFailedException()

        if data['result']:
            return True

        if data.get('reason') == 'unauthorized':
            raise AccessDeniedException()

        raise RequestFailedException(data.get('reason'))

    def _prepare_release(self, mod):
        meta = mod.get()

        for prop in ('screenshots', 'attachments'):
            sums = []
            for fn in meta[prop]:
                path = os.path.join(mod.folder, fn)

                if os.path.isfile(path):
                    chk = util.gen_hash(path)[1]

                    self.upload_file(fn, path)
                    sums.append(chk)

            meta[prop] = sums

        if meta['banner']:
            image = os.path.join(mod.folder, meta['banner'])

            if os.path.isfile(image):
                meta['banner'] = util.gen_hash(image)[1]
                self.upload_file('banner', image)

        return meta

    def create_release(self, mod):
        meta = self._prepare_release(mod)
        result = self._call('mod/release', check_code=True, json=meta)
        data = result.json()
        if not data:
            raise RequestFailedException('unknown')

        if data['result']:
            return True

        if data.get('reason') == 'unauthorized':
            raise AccessDeniedException(data['reason'])

        raise RequestFailedException(data.get('reason'))

    def update_release(self, mod):
        meta = self._prepare_release(mod)
        result = self._call('mod/release/update', check_code=True, json=meta)
        data = result.json()
        if not data:
            raise RequestFailedException('unknown')

        if data['result']:
            return True

        if data.get('reason') == 'unauthorized':
            raise AccessDeniedException(data['reason'])

        raise RequestFailedException(data.get('reason'))

    def report_release(self, mod, message):
        result = self._call('mod/release/report', check_code=True, data={
            'mid': mod.mid,
            'version': str(mod.version),
            'message': message
        })
        data = result.json()
        if not data or not data.get('result'):
            raise RequestFailedException('unknown')

        return True

    def delete_release(self, mod):
        result = self._call('mod/release/delete', check_code=True, data={
            'mid': mod.mid,
            'version': str(mod.version)
        })
        data = result.json()
        if not data or not data.get('result'):
            if data.get('reason') == 'unauthorized':
                raise AccessDeniedException(data['reason'])
            else:
                raise RequestFailedException('unknown')

        return True

    def upload_file(self, name, path, fn=None, content_checksum=None):
        _, checksum = util.gen_hash(path)

        result = self._call('upload/check', check_code=True, data={'checksum': checksum})
        data = result.json()
        if data.get('result'):
            # Already uploaded
            return True

        hdl = open(path, 'rb')
        enc = MultipartEncoder({
            'checksum': checksum,
            'content_checksum': content_checksum,
            'file': ('upload', hdl, 'application/octet-stream')
        })

        enc_len = enc.len

        def cb(monitor):
            progress.update(monitor.bytes_read / enc_len, 'Uploading %s...' % name)

        monitor = MultipartEncoderMonitor(enc, cb)
        self._call('upload/file', data=monitor, headers={
            'Content-Type': monitor.content_type
        }, check_code=True)

        return True

    def is_uploaded(self, checksum=None, content_checksum=None):
        assert checksum or content_checksum

        if checksum:
            data = {'checksum': checksum}
        else:
            data = {'content_checksum': content_checksum}

        result = self._call('upload/check', check_code=True, data=data)
        data = result.json()
        if data.get('result'):
            return True, data
        else:
            return False, None



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging


os.environ['DBUS_FATAL_WARNINGS'] = '0'

try:
    from PyQt5 import QtCore, QtGui, QtWidgets, QtNetwork

    try:
        from PyQt5 import QtWebChannel, QtWebEngineWidgets
    except ImportError:
        from PyQt5 import QtWebKit, QtWebKitWidgets

        QtWebChannel = None

        class QtWebEngineWidgets(object):
            QWebEngineView = QtWebKitWidgets.QWebView

    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
    QtCore.QString = str

    # Success!
    variant = 'PyQt5'

except ImportError:
    logging.exception('I was unable to load Qt! Tried PyQt5.')
    sys.exit(1)


class SignalContainer(QtCore.QObject):
    signal = QtCore.Signal(list)


def read_file(path, decode=True):
    fd = QtCore.QFile(path)
    if not fd.open(QtCore.QIODevice.ReadOnly):
        return None

    data = fd.readAll().data()
    fd.close()

    if decode:
        return data.decode('utf-8')
    else:
        return data


def load_styles(*names):
    data = ''
    for name in names:
        if ':/' in name:
            sheet = read_file(name)
        else:
            with open(name, 'r') as stream:
                sheet = stream.read()

        data += sheet.replace('url(./', 'url(' + os.path.dirname(name) + '/')

    return data


# This wrapper makes sure that the wrapped function is always run in the QT main thread.
def run_in_qt(func):
    cont = SignalContainer()

    def dispatcher(*args):
        cont.signal.emit(list(args))

    def listener(params):
        func(*params)

    cont.signal.connect(listener)

    return dispatcher


__all__ = [
    'QtCore', 'QtGui', 'QtWidgets', 'QtNetwork', 'QtWebChannel', 'QtWebEngineWidgets', 'QtWebKit', 'variant',
    'read_file', 'load_styles', 'run_in_qt'
]
